/*
 Name:		Fuzzy_Tank_V2_Library.cpp
 Created:	8/27/2017 2:27:47 AM
 Author:	georgychen
 Editor:	http://www.visualmicro.com
*/

#include "Fuzzy_Tank.h"
#include <EEPROM.h>

FuzzyTank::FuzzyTank()
{
	
}

#pragma region Setup and pairing
void FuzzyTank::begin()
{
	Serial << F("Initializing tank.") <<endl;
	Serial << F("Version ") << F(VERSION_NUMBER) << endl;
	userChannel = EEPROM.read(EEPROM_USER_CHANNEL);
	Serial << F("EEPROM Channel = ") << userChannel << endl;
	serialNumber = (uint16_t)EEPROM.read(EEPROM_SERIAL_NUMBER_HIGH) << 8 | (uint16_t)EEPROM.read(EEPROM_SERIAL_NUMBER_LOW);
	Serial << F("Device serial number = ") << serialNumber << endl;
	motorType = (MOTOR_TYPE)EEPROM.read(EEPROM_MOTOR_TYPE);
	Serial << F("Motor type = ");
	switch (motorType)
	{
		case DEFAULT_MOTOR:
		{
			Serial << F("default") << endl;
			break;
		}
		case TB16:
		{
			Serial << F("TB16") << endl;
			break;
		}
		case GA20:
		{
			Serial << F("TB16") << endl;
			break;
		}
		default:
		{
			Serial << F("undefined") << endl;
			break;
		}
	}
	gearsetType = (GEARSET_TYPE)EEPROM.read(EEPROM_GEARSET_TYPE);
	Serial << F("Gearset type = ");
	switch (gearsetType)
	{
		case DEFAULT_GEARSET:
		{
			Serial << F("default") << endl;
			break;
		}
		case TOWER_GEARSET_50_TO_1:
		{
			Serial << F("TOWER_GEARSET_50_TO_1") << endl;
			break;
		}
		case SY_PLANETARY_GEARSET_53_TO_1:
		{
			Serial << F("SY_PLANETARY_GEARSET_53_TO_1") << endl;
			break;
		}
		case SY_PLANETARY_GEARSET_29_TO_1:
		{
			Serial << F("SY_PLANETARY_GEARSET_29_TO_1") << endl;
			break;
		}
		default:
		{
			Serial << F("undefined") << endl;
			break;
		}
	}


	setPins();
	softSerialBegin();
	mp3Begin();
	motorBegin();
	frontLedBegin();
	ws2812Begin();
	RF24Begin();
	uint8_t tankAddress[] = { 51,52,53,54,55 };
	enableAckPayload();
	openReadingPipe(1, tankAddress);
	setDataRate(RF24_250KBPS);
	if (userChannel < 125 && userChannel > 0)
	{
		setChannel(userChannel);
	}
	startListening();

	if (EEPROM.read(EEPROM_USER_CHANNEL) == 119)
	{
		setEyesColor(fixColor[C_BLUE]);
		setHeadColor(fixColor[C_BLUE]);
		eyesColor.r = fixColor[C_BLUE][0];
		eyesColor.g = fixColor[C_BLUE][1];
		eyesColor.b = fixColor[C_BLUE][2];
		headColor.r = fixColor[C_BLUE][0];
		headColor.g = fixColor[C_BLUE][1];
		headColor.b = fixColor[C_BLUE][2];
	}
	if (EEPROM.read(EEPROM_USER_CHANNEL) == 118)
	{
		setEyesColor(fixColor[C_RED]);
		setHeadColor(fixColor[C_RED]);
		eyesColor.r = fixColor[C_RED][0];
		eyesColor.g = fixColor[C_RED][1];
		eyesColor.b = fixColor[C_RED][2];
		headColor.r = fixColor[C_RED][0];
		headColor.g = fixColor[C_RED][1];
		headColor.b = fixColor[C_RED][2];
	}
	if (EEPROM.read(EEPROM_USER_CHANNEL) == 117)
	{
		setEyesColor(fixColor[C_PURPLE]);
		setHeadColor(fixColor[C_PURPLE]);
		eyesColor.r = fixColor[C_PURPLE][0];
		eyesColor.g = fixColor[C_PURPLE][1];
		eyesColor.b = fixColor[C_PURPLE][2];
		headColor.r = fixColor[C_PURPLE][0];
		headColor.g = fixColor[C_PURPLE][1];
		headColor.b = fixColor[C_PURPLE][2];

	}
	if (EEPROM.read(EEPROM_USER_CHANNEL) == 116)
	{
		setEyesColor(fixColor[C_ORANGE]);
		setHeadColor(fixColor[C_ORANGE]);
		eyesColor.r = fixColor[C_ORANGE][0];
		eyesColor.g = fixColor[C_ORANGE][1];
		eyesColor.b = fixColor[C_ORANGE][2];
		headColor.r = fixColor[C_ORANGE][0];
		headColor.g = fixColor[C_ORANGE][1];
		headColor.b = fixColor[C_ORANGE][2];
	}
	if (EEPROM.read(EEPROM_USER_CHANNEL) == 115)
	{
		setEyesColor(fixColor[C_GREEN]);
		setHeadColor(fixColor[C_GREEN]);
		eyesColor.r = fixColor[C_GREEN][0];
		eyesColor.g = fixColor[C_GREEN][1];
		eyesColor.b = fixColor[C_GREEN][2];
		headColor.r = fixColor[C_GREEN][0];
		headColor.g = fixColor[C_GREEN][1];
		headColor.b = fixColor[C_GREEN][2];
	}
	eyesLedIsOn = true;
	updateColor();
	

	servoBegin();

	Serial << F("Initializing IR receiver interrupt.") << endl;
	attachInterrupt(digitalPinToInterrupt(IRReceiverPin), IRLinterrupt<IR_ALL>, CHANGE);
	irSendPacket.NEC_address.byte1 = 1;
	irSendPacket.NEC_address.byte2 = 101;
	irSendPacket.NEC_command.byte3 = 201;
	irSendPacket.NEC_command.commandNot = ~irSendPacket.NEC_command.byte3;

	setFrontLedDimOn();
	Serial << F("Set servo1 to retracted position.") << endl;
	setServo1AndOff(SERVO1_RETRACTED_POSITION, 1000);

	setDevelopmentBoardDifference();
	Serial << F("Initialization completed.") << endl;
}

void FuzzyTank::begin(uint8_t userAssignedBoardVersion)
{
	EEPROM.update(EEPROM_BOARD_VERSION, userAssignedBoardVersion);
	begin();
}

void FuzzyTank::setPins()
{
	//read boardVersion from EEPROM
	boardVersion = EEPROM.read(EEPROM_BOARD_VERSION);
	
	Serial << F("EEPROM Board version = ") << boardVersion << endl;
	bool isSupported = false;
	for (uint8_t i = 0; i < SUPPORTED_BOARD_TYPES; i++)
	{
		if (boardVersion == validBoardVersions[i])
		{
			isSupported = true;
		}
	}
	if (isSupported == true)
	{
		Serial << F("Board version is supported.") << endl;
	}
	else
	{
		Serial << F("Board version is not supported, program halted.") << endl;

		while (true);
	}
	switch (boardVersion)
	{
		case 2:
		{
			yx5200SerialPin = YX5200_SERIAL_PIN_V2;
			motorAPwmPin = MOTOR_A_PWM_PIN_V2;
			motorAIn1Pin = MOTOR_A_IN1_PIN_V2;
			motorAIn2Pin = MOTOR_A_IN2_PIN_V2;
			motorBPwmPin = MOTOR_B_PWM_PIN_V2;
			motorBIn1Pin = MOTOR_B_IN1_PIN_V2;
			motorBIn2Pin = MOTOR_B_IN2_PIN_V2;
			RF24CsnPin = RF24_CSN_PIN_V2;
			RF24CePin = RF24_CE_PIN_V2;
			IRReceiverPin = IR_RECEIVER_PIN_V2;
			IRSenderPin = IR_SENDER_PIN_V2;
			WS2812LedPin = WS2812_LED_PIN_V2;
			frontLedPin = FRONT_LED_PIN_V2;
			break;
		}
		case 5:
		{
			yx5200SerialPin = YX5200_SERIAL_PIN_V5;
			motorAPwmPin = MOTOR_A_PWM_PIN_V5;
			motorAIn1Pin = MOTOR_A_IN1_PIN_V5;
			motorAIn2Pin = MOTOR_A_IN2_PIN_V5;
			motorBPwmPin = MOTOR_B_PWM_PIN_V5;
			motorBIn1Pin = MOTOR_B_IN1_PIN_V5;
			motorBIn2Pin = MOTOR_B_IN2_PIN_V5;
			RF24CsnPin = RF24_CSN_PIN_V5;
			RF24CePin = RF24_CE_PIN_V5;
			IRReceiverPin = IR_RECEIVER_PIN_V5;
			IRSenderPin = IR_SENDER_PIN_V5;
			WS2812LedPin = WS2812_LED_PIN_V5;
			frontLedPin = FRONT_LED_PIN_V5;
			break;
		}
		case 6:
		{
			yx5200SerialPin = YX5200_SERIAL_PIN_V6;
			motorAPwmPin = MOTOR_A_PWM_PIN_V6;
			motorAIn1Pin = MOTOR_A_IN1_PIN_V6;
			motorAIn2Pin = MOTOR_A_IN2_PIN_V6;
			motorBPwmPin = MOTOR_B_PWM_PIN_V6;
			motorBIn1Pin = MOTOR_B_IN1_PIN_V6;
			motorBIn2Pin = MOTOR_B_IN2_PIN_V6;
			RF24CsnPin = RF24_CSN_PIN_V6;
			RF24CePin = RF24_CE_PIN_V6;
			IRReceiverPin = IR_RECEIVER_PIN_V6;
			IRSenderPin = IR_SENDER_PIN_V6;
			WS2812LedPin = WS2812_LED_PIN_V6;
			frontLedPin = FRONT_LED_PIN_V6;
			break;
		}
		case 7:
		{
			yx5200SerialPin = YX5200_SERIAL_PIN_V7;
			motorAPwmPin = MOTOR_A_PWM_PIN_V7;
			motorAIn1Pin = MOTOR_A_IN1_PIN_V7;
			motorAIn2Pin = MOTOR_A_IN2_PIN_V7;
			motorBPwmPin = MOTOR_B_PWM_PIN_V7;
			motorBIn1Pin = MOTOR_B_IN1_PIN_V7;
			motorBIn2Pin = MOTOR_B_IN2_PIN_V7;
			RF24CsnPin = RF24_CSN_PIN_V7;
			RF24CePin = RF24_CE_PIN_V7;
			IRReceiverPin = IR_RECEIVER_PIN_V7;
			IRSenderPin = IR_SENDER_PIN_V7;
			WS2812LedPin = WS2812_LED_PIN_V7;
			frontLedPin = FRONT_LED_PIN_V7;
			break;
		}
	}


	
}

void FuzzyTank::setDevelopmentBoardDifference()
{
	switch (serialNumber)
	{
		case 1: //increase speed due to smaller motor
		{
			//maxForwardSpeed = 255;
			//maxBackwardSpeed = 255;
			//motorPWMLowThreshold = 100;
			//maxTurnSpeed = 72;
			//reviveMotorCommandValue = -100;
			break;
		}
	}
}

void FuzzyTank::autoPairing()
{
	Serial << F("Begin auto-pairing.") << endl;

	Serial << F("Auto-pairing ended.") << endl;
}

void FuzzyTank::setMotorType(MOTOR_TYPE motor)
{
	EEPROM.update(EEPROM_MOTOR_TYPE, motor);
}

void FuzzyTank::setGearsetType(GEARSET_TYPE gearset)
{
	EEPROM.update(EEPROM_GEARSET_TYPE, gearset);
}

#pragma endregion Setup and pairing

#pragma region MP3
void FuzzyTank::playMp3(uint16_t fileNumber)
{
	mp3TxBuffer[3] = 0x12;
	mp3TxBuffer[5] = (uint8_t)(fileNumber >> 8);
	mp3TxBuffer[6] = (uint8_t)(fileNumber);
	sendBufferToMp3();
}

void FuzzyTank::update()
{
	if (isGameOver == false)
	{
		updateServo();
		updateRF24();
		updateFrontLed();
		updateIRReceiver();
		updateRGBLed();
		updateGamePlay();
	}
	else //isGameOver == true
	{
		updateGamePlay();
		updateIRReceiver(); //can get hit and complain after game over!
	}
}

void FuzzyTank::softSerialBegin()
{
	Serial << F("Initializing software serial.") << endl;
	_transmitBitMask = digitalPinToBitMask(yx5200SerialPin);
	digitalWrite(yx5200SerialPin, _inverse_logic ? LOW : HIGH);
	pinMode(yx5200SerialPin, OUTPUT);
	_transmitBitMask = digitalPinToBitMask(yx5200SerialPin);
	uint8_t port = digitalPinToPort(yx5200SerialPin);
	_transmitPortRegister = portOutputRegister(port);
}

void FuzzyTank::mp3Begin()
{
	Serial << F("Initializing mp3.") << endl;
	mp3TxBuffer[0] = 0x7E; //required start byte 0x7E
	mp3TxBuffer[1] = 0xFF; //version byte
	mp3TxBuffer[2] = 0x06; //packet length
	mp3TxBuffer[3] = 0x0C; //command byte, 0x0C = reset
	mp3TxBuffer[4] = 0x00; //feedback
	mp3TxBuffer[5] = 0x00; //parameter byte 1
	mp3TxBuffer[6] = 0x00; //parameter byte 2
	mp3TxBuffer[7] = 0x00; //check sum
	mp3TxBuffer[8] = 0x00; //check sum
	mp3TxBuffer[9] = 0xEF; //required end byte 0xEF

	sendBufferToMp3(); //send a reset command
	delay(1000);//required after any reset
	setVolume(MP3_DEFAULT_VOLUME);
	delay(100);
}

void FuzzyTank::writeChecksum()
{
	uint16_t checksum = 0;
	for (uint8_t i = 1; i < 7; i++)
	{
		checksum += mp3TxBuffer[i];
	}
	checksum = -checksum;
	mp3TxBuffer[7] = (uint8_t)(checksum >> 8);
	mp3TxBuffer[8] = (uint8_t)(checksum);
}

void FuzzyTank::sendBufferToMp3()
{
	writeChecksum();
	for (uint8_t i = 0; i < 10; i++)
	{
		serialWrite(mp3TxBuffer[i]);
	}
}

void FuzzyTank::setVolume(uint8_t vol)
{
	mp3TxBuffer[3] = 0x06;
	mp3TxBuffer[5] = (uint8_t)(vol >> 8);
	mp3TxBuffer[6] = (uint8_t)(vol);
	sendBufferToMp3();
}

inline void FuzzyTank::tunedDelay(uint16_t delay)
{
	_delay_loop_2(delay);
}

void FuzzyTank::serialWrite(uint8_t b)
{
	// By declaring these as local variables, the compiler will put them
	// in registers _before_ disabling interrupts and entering the
	// critical timing sections below, which makes it a lot easier to
	// verify the cycle timings
	volatile uint8_t *reg = _transmitPortRegister;
	uint8_t reg_mask = _transmitBitMask;
	uint8_t inv_mask = ~_transmitBitMask;
	uint8_t oldSREG = SREG;
	bool inv = false;
	uint16_t delay = _tx_delay;

	if (inv)
		b = ~b;

	cli();  // turn off interrupts for a clean txmit

			// Write the start bit
	if (inv)
		*reg |= reg_mask;
	else
		*reg &= inv_mask;

	tunedDelay(delay);

	// Write each of the 8 bits
	for (uint8_t i = 8; i > 0; --i)
	{
		if (b & 1) // choose bit
			*reg |= reg_mask; // send 1
		else
			*reg &= inv_mask; // send 0

		tunedDelay(delay);
		b >>= 1;
	}

	// restore pin to natural state
	if (inv)
		*reg &= inv_mask;
	else
		*reg |= reg_mask;

	SREG = oldSREG; // turn interrupts back on
	tunedDelay(_tx_delay);
}
#pragma endregion MP3

#pragma region WS2812 led
void FuzzyTank::ws2812Begin()
{
	Serial << F("Initializing ws2812 LED.") << endl;
	pinMask = digitalPinToBitMask(WS2812LedPin);
	ws2812_port = portOutputRegister(digitalPinToPort(WS2812LedPin));
	ws2812_port_reg = portModeRegister(digitalPinToPort(WS2812LedPin));
}

void FuzzyTank::setColor(uint8_t pixelIndex, colorRGB color)
{
	uint8_t pixelPosition = pixelIndex * 3;
	pixels[pixelPosition] = color.g;
	pixels[pixelPosition + 1] = color.r;
	pixels[pixelPosition + 2] = color.b;
}

void FuzzyTank::setColor(uint8_t pixelIndex, uint8_t * color)
{
	uint8_t pixelPosition = pixelIndex * 3;
	pixels[pixelPosition] = color[1];
	pixels[pixelPosition + 1] = color[0];
	pixels[pixelPosition + 2] = color[+2];
}

void FuzzyTank::setEyesColor(uint8_t * color)
{
	setColor(0, color);
	setColor(1, color);
}

void FuzzyTank::setEyesColor(colorRGB color)
{
	setColor(0, color);
	setColor(1, color);
}

void FuzzyTank::setHeadColor(uint8_t * color)
{
	setColor(2, color);
	setColor(3, color);
	setColor(4, color);
}

void FuzzyTank::setHeadColor(colorRGB color)
{
	setColor(2, color);
	setColor(3, color);
	setColor(4, color);
}

void FuzzyTank::updateColor()
{
	*ws2812_port_reg |= pinMask; // Enable DDR
	ws2812_sendarray_mask(pixels, 3 * WS2812_LED_COUNT, pinMask, (uint8_t*)ws2812_port, (uint8_t*)ws2812_port_reg);
}

void FuzzyTank::ws2812_sendarray_mask(uint8_t *data, uint16_t datlen, uint8_t maskhi, uint8_t *port, uint8_t *portreg)
{
	uint8_t curbyte, ctr, masklo;
	uint8_t sreg_prev;

	masklo = ~maskhi & *port;
	maskhi |= *port;
	sreg_prev = SREG;
	cli();

	while (datlen--) {
		curbyte = *data++;

		asm volatile(
			"       ldi   %0,8  \n\t"
			"loop%=:            \n\t"
			"       st    X,%3 \n\t"    //  '1' [02] '0' [02] - re
#if (w1_nops&1)
			w_nop1
#endif
#if (w1_nops&2)
			w_nop2
#endif
#if (w1_nops&4)
			w_nop4
#endif
#if (w1_nops&8)
			w_nop8
#endif
#if (w1_nops&16)
			w_nop16
#endif
			"       sbrs  %1,7  \n\t"    //  '1' [04] '0' [03]
			"       st    X,%4 \n\t"     //  '1' [--] '0' [05] - fe-low
			"       lsl   %1    \n\t"    //  '1' [05] '0' [06]
#if (w2_nops&1)
			w_nop1
#endif
#if (w2_nops&2)
			w_nop2
#endif
#if (w2_nops&4)
			w_nop4
#endif
#if (w2_nops&8)
			w_nop8
#endif
#if (w2_nops&16)
			w_nop16
#endif
			"       brcc skipone%= \n\t"    //  '1' [+1] '0' [+2] - 
			"       st   X,%4      \n\t"    //  '1' [+3] '0' [--] - fe-high
			"skipone%=:               "     //  '1' [+3] '0' [+2] - 

#if (w3_nops&1)
			w_nop1
#endif
#if (w3_nops&2)
			w_nop2
#endif
#if (w3_nops&4)
			w_nop4
#endif
#if (w3_nops&8)
			w_nop8
#endif
#if (w3_nops&16)
			w_nop16
#endif

			"       dec   %0    \n\t"    //  '1' [+4] '0' [+3]
			"       brne  loop%=\n\t"    //  '1' [+5] '0' [+4]
			:	"=&d" (ctr)
			//    :	"r" (curbyte), "I" (_SFR_IO_ADDR(ws2812_PORTREG)), "r" (maskhi), "r" (masklo)
			: "r" (curbyte), "x" (port), "r" (maskhi), "r" (masklo)
			);
	}
	SREG = sreg_prev;
}

void FuzzyTank::updateRGBLed()
{
	if (blinkEyes == true)
	{
		if (eyesLedIsOn == true)
		{
			if (millis() - blinkEyesTimer > blinkEyesTimeDuration)
			{
				colorRGB dark;
				uint8_t darkness = random(3, 5);
				dark.r = eyesColor.r>> darkness;
				dark.g = eyesColor.g>> darkness;
				dark.b = eyesColor.b>> darkness;
				setEyesColor(dark);
				updateColor();
				blinkEyesTimeDuration = random(100, 300);
				eyesLedIsOn = false;
				blinkEyesTimer = millis();
			}
		}
		else //(eyesLedIsOn == false)
		{
			if (millis() - blinkEyesTimer > blinkEyesTimeDuration)
			{
				setEyesColor(eyesColor);
				updateColor();
				blinkEyesTimeDuration = random(2000, 5000);
				eyesLedIsOn = true;
				blinkEyesTimer = millis();
			}
		}
	}

	if (isInvincible == true)
	{

	}
	else
	{
		if (flashHeadLed == true)
		{
			if (headLedIsOn == true)
			{
				if (millis() - bloodStateFlashingTimer > bloodStateFlashingDuration)
				{
					colorRGB dark;
					dark.r = headColor.r >> 3;
					dark.g = headColor.g >> 3;
					dark.b = headColor.b >> 3;
					setHeadColor(dark);
					updateColor();

					headLedIsOn = false;
					bloodStateFlashingTimer = millis();
				}
			}
			else //(headLedIsOn == false)
			{
				if (millis() - bloodStateFlashingTimer > bloodStateFlashingDuration)
				{
					setHeadColor(headColor);
					updateColor();
					headLedIsOn = true;
					bloodStateFlashingTimer = millis();
				}
			}
		}

	}
	
}
#pragma endregion WS2812

#pragma region Front LED
void FuzzyTank::frontLedBegin()
{
	Serial << F("Initializing front LED.") << endl;
	pinMode(frontLedPin, OUTPUT);
	digitalWrite(frontLedPin, LOW);
}

void FuzzyTank::frontLedOn(uint32_t time)
{
	digitalWrite(frontLedPin, HIGH);
	frontLedTimerEnabled = true;
	frontLedDuration = time;
	frontLedOnTimer = millis();
}

void FuzzyTank::updateFrontLed()
{
	switch (boardVersion)
	{
		case 2: //non-pwm pin
		case 5:
		case 6:
		case 7:
		{
			if (frontLedTimerEnabled == true)
			{
				if (millis() - frontLedOnTimer> frontLedDuration)
				{
					frontLedTimerEnabled = false;
					digitalWrite(frontLedPin, LOW);
				}
			}
			else  if (frontLedPWMEnabled == true)  //(frontLedTimerEnabled == false)
			{
				{
					if (millis() - frontLedPwmTimer > FRONT_LED_PWM_OFF_PERIOD)
					{
						digitalWrite(frontLedPin, HIGH);
						delayMicroseconds(100);
						digitalWrite(frontLedPin, LOW);
						frontLedPwmTimer = millis();
					}
				}
			}
			break;
		}
		case 0: //on pwm pin d10
		{
			break;
		}

	}
	
}

void FuzzyTank::setFrontLedDimOn()
{
	switch (boardVersion)
	{
		case 2: //non-pwm pin
		case 5:
		case 6:
		case 7:
		{
			frontLedPWMEnabled = true;
			break;
		}
		case 0: //on pwm pin d10
		{
			analogWrite(frontLedPin, 240);
			break;
		}
	}
}

void FuzzyTank::setFrontLedDimOff()
{
	switch (boardVersion)
	{
		case 2: //non-pwm pin
		case 5:
		case 6:
		case 7:

		{
			frontLedPWMEnabled = false;
			break;
		}
		case 0: //on pwm pin d10
		{
			analogWrite(frontLedPin, 0);
			break;
		}
	}
}
#pragma endregion Front LED

#pragma region Servo
void FuzzyTank::servoBegin()
{
	Serial << F("Initializing servo.") << endl;
	TCCR1A = _BV(WGM11);							// Mode 14 (fast PWM)
	TCCR1B = _BV(WGM13) | _BV(WGM12) | _BV(CS11);	// 1:8 prescale
	ICR1 = F_CPU / 8 / 50;							// ~50 Hz (~20 ms)
	OCR1A = 1500 * (F_CPU / 1000000L) / 8;			// Center servos
	if (TCNT1 > ICR1) TCNT1 = 0;                    // Restart count
}

void FuzzyTank::setServo1(uint16_t position)
{
	OCR1A = position* (F_CPU / 1000000L) / 8;
}

void FuzzyTank::enableServo1()
{
	TCCR1A |= _BV(COM1A1);	//turn on Fast PWM
	DDRB |= _BV(PB1);		//enable output
	servo1IsEnabled = true;
}

void FuzzyTank::disableServo1()
{
	while (TCNT1 < OCR1A);
	TCCR1A &= ~_BV(COM1A1);	//turn off Fast PWM
	DDRB &= ~_BV(PB1);		//disable output
	servo1IsEnabled = false;
}

void FuzzyTank::updateServo()
{
	if (servo1OffTimerIsEnabled == true)
	{
		if (millis() - servo1OffTimer > servo1OffPeroid)
		{
			disableServo1();
			servo1OffTimerIsEnabled = false;
		}
	}
}

void FuzzyTank::setServo1AndOff(uint16_t position, uint32_t period)
{
	if (servo1IsEnabled == false) enableServo1();
	setServo1(position);
	servo1OffTimerIsEnabled = true;
	servo1OffTimer = millis();
	servo1OffPeroid = period;
}

#pragma endregion Servo

#pragma region RF24

void FuzzyTank::csn(bool mode)
{

#if defined (RF24_TINY)
	if (ce_pin != csn_pin) {
		digitalWrite(csn_pin, mode);
	}
	else {
		if (mode == HIGH) {
			PORTB |= (1 << PINB2);  	// SCK->CSN HIGH
			delayMicroseconds(100); // allow csn to settle.
		}
		else {
			PORTB &= ~(1 << PINB2);	// SCK->CSN LOW
			delayMicroseconds(11);  // allow csn to settle
		}
	}
	// Return, CSN toggle complete
	return;

#elif defined(ARDUINO) && !defined (RF24_SPI_TRANSACTIONS)
	// Minimum ideal SPI bus speed is 2x data rate
	// If we assume 2Mbs data rate and 16Mhz clock, a
	// divider of 4 is the minimum we want.
	// CLK:BUS 8Mhz:2Mhz, 16Mhz:4Mhz, or 20Mhz:5Mhz

#if !defined (SOFTSPI)	
	_SPI.setBitOrder(MSBFIRST);
	_SPI.setDataMode(SPI_MODE0);
	_SPI.setClockDivider(SPI_CLOCK_DIV2);
#endif
#elif defined (RF24_RPi)
	if (!mode)
		_SPI.chipSelect(csn_pin);
#endif

#if !defined (RF24_LINUX)
	digitalWrite(csn_pin, mode);
	delayMicroseconds(csDelay);
#endif

}

/****************************************************************************/

void FuzzyTank::ce(bool level)
{
	//Allow for 3-pin use on ATTiny
	if (ce_pin != csn_pin) digitalWrite(ce_pin, level);
}

/****************************************************************************/

inline void FuzzyTank::beginTransaction() {
#if defined (RF24_SPI_TRANSACTIONS)
	_SPI.beginTransaction(SPISettings(RF24_SPI_SPEED, MSBFIRST, SPI_MODE0));
#endif
	csn(LOW);
}

/****************************************************************************/

inline void FuzzyTank::endTransaction() {
	csn(HIGH);
#if defined (RF24_SPI_TRANSACTIONS)
	_SPI.endTransaction();
#endif
}

/****************************************************************************/

uint8_t FuzzyTank::read_register(uint8_t reg, uint8_t* buf, uint8_t len)
{
	uint8_t status;

#if defined (RF24_LINUX)
	beginTransaction(); //configures the spi settings for RPi, locks mutex and setting csn low
	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	uint8_t size = len + 1; // Add register value to transmit buffer

	*ptx++ = (R_REGISTER | (REGISTER_MASK & reg));

	while (len--) { *ptx++ = NOP; } // Dummy operation, just for reading

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, size);

	status = *prx++; // status is 1st byte of receive buffer

					 // decrement before to skip status byte
	while (--size) { *buf++ = *prx++; }
	endTransaction(); //unlocks mutex and setting csn high

#else

	beginTransaction();
	status = _SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
	while (len--) {
		*buf++ = _SPI.transfer(0xff);
	}
	endTransaction();

#endif

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::read_register(uint8_t reg)
{
	uint8_t result;

#if defined (RF24_LINUX)

	beginTransaction();

	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	*ptx++ = (R_REGISTER | (REGISTER_MASK & reg));
	*ptx++ = NOP; // Dummy operation, just for reading

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, 2);
	result = *++prx;   // result is 2nd byte of receive buffer

	endTransaction();
#else

	beginTransaction();
	_SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
	result = _SPI.transfer(0xff);
	endTransaction();

#endif

	return result;
}

/****************************************************************************/

uint8_t FuzzyTank::write_register(uint8_t reg, const uint8_t* buf, uint8_t len)
{
	uint8_t status;

#if defined (RF24_LINUX) 
	beginTransaction();
	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	uint8_t size = len + 1; // Add register value to transmit buffer

	*ptx++ = (W_REGISTER | (REGISTER_MASK & reg));
	while (len--)
		*ptx++ = *buf++;

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, size);
	status = *prx; // status is 1st byte of receive buffer
	endTransaction();
#else

	beginTransaction();
	status = _SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
	while (len--)
		_SPI.transfer(*buf++);
	endTransaction();

#endif

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::write_register(uint8_t reg, uint8_t value)
{
	uint8_t status;

	IF_SERIAL_DEBUG(printf_P(PSTR("write_register(%02x,%02x)\r\n"), reg, value));

#if defined (RF24_LINUX)
	beginTransaction();
	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	*ptx++ = (W_REGISTER | (REGISTER_MASK & reg));
	*ptx = value;

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, 2);
	status = *prx++; // status is 1st byte of receive buffer
	endTransaction();
#else

	beginTransaction();
	status = _SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
	_SPI.transfer(value);
	endTransaction();

#endif

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::write_payload(const void* buf, uint8_t data_len, const uint8_t writeType)
{
	uint8_t status;
	const uint8_t* current = reinterpret_cast<const uint8_t*>(buf);

	data_len = rf24_min(data_len, payload_size);
	uint8_t blank_len = dynamic_payloads_enabled ? 0 : payload_size - data_len;

	//printf("[Writing %u bytes %u blanks]",data_len,blank_len);
	IF_SERIAL_DEBUG(printf("[Writing %u bytes %u blanks]\n", data_len, blank_len); );

#if defined (RF24_LINUX)
	beginTransaction();
	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	uint8_t size;
	size = data_len + blank_len + 1; // Add register value to transmit buffer

	*ptx++ = writeType;
	while (data_len--)
		*ptx++ = *current++;
	while (blank_len--)
		*ptx++ = 0;

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, size);
	status = *prx; // status is 1st byte of receive buffer
	endTransaction();

#else

	beginTransaction();
	status = _SPI.transfer(writeType);
	while (data_len--) {
		_SPI.transfer(*current++);
	}
	while (blank_len--) {
		_SPI.transfer(0);
	}
	endTransaction();

#endif

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::read_payload(void* buf, uint8_t data_len)
{
	uint8_t status;
	uint8_t* current = reinterpret_cast<uint8_t*>(buf);

	if (data_len > payload_size) data_len = payload_size;
	uint8_t blank_len = dynamic_payloads_enabled ? 0 : payload_size - data_len;

	//printf("[Reading %u bytes %u blanks]",data_len,blank_len);

	IF_SERIAL_DEBUG(printf("[Reading %u bytes %u blanks]\n", data_len, blank_len); );

#if defined (RF24_LINUX)
	beginTransaction();
	uint8_t * prx = spi_rxbuff;
	uint8_t * ptx = spi_txbuff;
	uint8_t size;
	size = data_len + blank_len + 1; // Add register value to transmit buffer

	*ptx++ = R_RX_PAYLOAD;
	while (--size)
		*ptx++ = NOP;

	size = data_len + blank_len + 1; // Size has been lost during while, re affect

	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, size);

	status = *prx++; // 1st byte is status	

	if (data_len > 0) {
		while (--data_len) // Decrement before to skip 1st status byte
			*current++ = *prx++;

		*current = *prx;
	}
	endTransaction();
#else

	beginTransaction();
	status = _SPI.transfer(R_RX_PAYLOAD);
	while (data_len--) {
		*current++ = _SPI.transfer(0xFF);
	}
	while (blank_len--) {
		_SPI.transfer(0xff);
	}
	endTransaction();

#endif

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::flush_rx(void)
{
	return spiTrans(FLUSH_RX);
}

/****************************************************************************/

uint8_t FuzzyTank::flush_tx(void)
{
	return spiTrans(FLUSH_TX);
}

/****************************************************************************/

uint8_t FuzzyTank::spiTrans(uint8_t cmd) {

	uint8_t status;

	beginTransaction();
	status = _SPI.transfer(cmd);
	endTransaction();

	return status;
}

/****************************************************************************/

uint8_t FuzzyTank::get_status(void)
{
	return spiTrans(NOP);
}

/****************************************************************************/
#if !defined (MINIMAL)
void FuzzyTank::print_status(uint8_t status)
{
	printf_P(PSTR("STATUS\t\t = 0x%02x RX_DR=%x TX_DS=%x MAX_RT=%x RX_P_NO=%x TX_FULL=%x\r\n"),
		status,
		(status & _BV(RX_DR)) ? 1 : 0,
		(status & _BV(TX_DS)) ? 1 : 0,
		(status & _BV(MAX_RT)) ? 1 : 0,
		((status >> RX_P_NO) & 0x07),
		(status & _BV(TX_FULL)) ? 1 : 0
	);
}

/****************************************************************************/

void FuzzyTank::print_observe_tx(uint8_t value)
{
	printf_P(PSTR("OBSERVE_TX=%02x: POLS_CNT=%x ARC_CNT=%x\r\n"),
		value,
		(value >> PLOS_CNT) & 0x0F,
		(value >> ARC_CNT) & 0x0F
	);
}

/****************************************************************************/

void FuzzyTank::print_byte_register(const char* name, uint8_t reg, uint8_t qty)
{
	//char extra_tab = strlen_P(name) < 8 ? '\t' : 0;
	//printf_P(PSTR(PRIPSTR"\t%c ="),name,extra_tab);
#if defined (RF24_LINUX)
	printf("%s\t =", name);
#else
	printf_P(PSTR(PRIPSTR"\t ="), name);
#endif
	while (qty--)
		printf_P(PSTR(" 0x%02x"), read_register(reg++));
	printf_P(PSTR("\r\n"));
}

/****************************************************************************/

void FuzzyTank::print_address_register(const char* name, uint8_t reg, uint8_t qty)
{

#if defined (RF24_LINUX)
	printf("%s\t =", name);
#else
	printf_P(PSTR(PRIPSTR"\t ="), name);
#endif
	while (qty--)
	{
		uint8_t buffer[addr_width];
		read_register(reg++, buffer, sizeof buffer);

		printf_P(PSTR(" 0x"));
		uint8_t* bufptr = buffer + sizeof buffer;
		while (--bufptr >= buffer)
			printf_P(PSTR("%02x"), *bufptr);
	}

	printf_P(PSTR("\r\n"));
}
#endif
/****************************************************************************/

FuzzyTank::FuzzyTank(uint16_t _cepin, uint16_t _cspin) :
	ce_pin(_cepin), csn_pin(_cspin), p_variant(false),
	payload_size(32), dynamic_payloads_enabled(false), addr_width(5), csDelay(5)//,pipe0_reading_address(0)
{
	pipe0_reading_address[0] = 0;
}

/****************************************************************************/

#if defined (RF24_LINUX) && !defined (MRAA)//RPi constructor

FuzzyTank::RF24(uint16_t _cepin, uint16_t _cspin, uint32_t _spi_speed) :
	ce_pin(_cepin), csn_pin(_cspin), spi_speed(_spi_speed), p_variant(false), payload_size(32), dynamic_payloads_enabled(false), addr_width(5)//,pipe0_reading_address(0) 
{
	pipe0_reading_address[0] = 0;
}
#endif

/****************************************************************************/

void FuzzyTank::setChannel(uint8_t channel)
{
	const uint8_t max_channel = 125;
	write_register(RF_CH, rf24_min(channel, max_channel));
}

uint8_t FuzzyTank::getChannel()
{

	return read_register(RF_CH);
}
/****************************************************************************/

void FuzzyTank::setPayloadSize(uint8_t size)
{
	payload_size = rf24_min(size, 32);
}

/****************************************************************************/

uint8_t FuzzyTank::getPayloadSize(void)
{
	//payload_size = read_register(RX_PW_P1);
	return payload_size;
}

/****************************************************************************/

#if !defined (MINIMAL)

static const char rf24_datarate_e_str_0[] PROGMEM = "1MBPS";
static const char rf24_datarate_e_str_1[] PROGMEM = "2MBPS";
static const char rf24_datarate_e_str_2[] PROGMEM = "250KBPS";
static const char * const rf24_datarate_e_str_P[] PROGMEM = {
	rf24_datarate_e_str_0,
	rf24_datarate_e_str_1,
	rf24_datarate_e_str_2,
};
static const char rf24_model_e_str_0[] PROGMEM = "nRF24L01";
static const char rf24_model_e_str_1[] PROGMEM = "nRF24L01+";
static const char * const rf24_model_e_str_P[] PROGMEM = {
	rf24_model_e_str_0,
	rf24_model_e_str_1,
};
static const char rf24_crclength_e_str_0[] PROGMEM = "Disabled";
static const char rf24_crclength_e_str_1[] PROGMEM = "8 bits";
static const char rf24_crclength_e_str_2[] PROGMEM = "16 bits";
static const char * const rf24_crclength_e_str_P[] PROGMEM = {
	rf24_crclength_e_str_0,
	rf24_crclength_e_str_1,
	rf24_crclength_e_str_2,
};
static const char rf24_pa_dbm_e_str_0[] PROGMEM = "PA_MIN";
static const char rf24_pa_dbm_e_str_1[] PROGMEM = "PA_LOW";
static const char rf24_pa_dbm_e_str_2[] PROGMEM = "PA_HIGH";
static const char rf24_pa_dbm_e_str_3[] PROGMEM = "PA_MAX";
static const char * const rf24_pa_dbm_e_str_P[] PROGMEM = {
	rf24_pa_dbm_e_str_0,
	rf24_pa_dbm_e_str_1,
	rf24_pa_dbm_e_str_2,
	rf24_pa_dbm_e_str_3,
};

#if defined (RF24_LINUX)
static const char rf24_csn_e_str_0[] = "CE0 (PI Hardware Driven)";
static const char rf24_csn_e_str_1[] = "CE1 (PI Hardware Driven)";
static const char rf24_csn_e_str_2[] = "CE2 (PI Hardware Driven)";
static const char rf24_csn_e_str_3[] = "Custom GPIO Software Driven";
static const char * const rf24_csn_e_str_P[] = {
	rf24_csn_e_str_0,
	rf24_csn_e_str_1,
	rf24_csn_e_str_2,
	rf24_csn_e_str_3,
};
#endif

void FuzzyTank::printDetails(void)
{

#if defined (RF24_RPi)
	printf("================ SPI Configuration ================\n");
	if (csn_pin < BCM2835_SPI_CS_NONE) {
		printf("CSN Pin  \t = %s\n", rf24_csn_e_str_P[csn_pin]);
	}
	else {
		printf("CSN Pin  \t = Custom GPIO%d%s\n", csn_pin,
			csn_pin == RPI_V2_GPIO_P1_26 ? " (CE1) Software Driven" : "");
	}
	printf("CE Pin  \t = Custom GPIO%d\n", ce_pin);
	printf("Clock Speed\t = ");
	switch (spi_speed)
	{
	case BCM2835_SPI_SPEED_64MHZ: printf("64 Mhz");	break;
	case BCM2835_SPI_SPEED_32MHZ: printf("32 Mhz");	break;
	case BCM2835_SPI_SPEED_16MHZ: printf("16 Mhz");	break;
	case BCM2835_SPI_SPEED_8MHZ: printf("8 Mhz");	break;
	case BCM2835_SPI_SPEED_4MHZ: printf("4 Mhz");	break;
	case BCM2835_SPI_SPEED_2MHZ: printf("2 Mhz");	break;
	case BCM2835_SPI_SPEED_1MHZ: printf("1 Mhz");	break;
	case BCM2835_SPI_SPEED_512KHZ: printf("512 KHz");	break;
	case BCM2835_SPI_SPEED_256KHZ: printf("256 KHz");	break;
	case BCM2835_SPI_SPEED_128KHZ: printf("128 KHz");	break;
	case BCM2835_SPI_SPEED_64KHZ: printf("64 KHz");	break;
	case BCM2835_SPI_SPEED_32KHZ: printf("32 KHz");	break;
	case BCM2835_SPI_SPEED_16KHZ: printf("16 KHz");	break;
	case BCM2835_SPI_SPEED_8KHZ: printf("8 KHz");	break;
	default: printf("8 Mhz");	break;
	}
	printf("\n================ NRF Configuration ================\n");

#endif //Linux

	print_status(get_status());

	print_address_register(PSTR("RX_ADDR_P0-1"), RX_ADDR_P0, 2);
	print_byte_register(PSTR("RX_ADDR_P2-5"), RX_ADDR_P2, 4);
	print_address_register(PSTR("TX_ADDR\t"), TX_ADDR);

	print_byte_register(PSTR("RX_PW_P0-6"), RX_PW_P0, 6);
	print_byte_register(PSTR("EN_AA\t"), EN_AA);
	print_byte_register(PSTR("EN_RXADDR"), EN_RXADDR);
	print_byte_register(PSTR("RF_CH\t"), RF_CH);
	print_byte_register(PSTR("RF_SETUP"), RF_SETUP);
	print_byte_register(PSTR("CONFIG\t"), NRF_CONFIG);
	print_byte_register(PSTR("DYNPD/FEATURE"), DYNPD, 2);

	printf_P(PSTR("Data Rate\t = " PRIPSTR "\r\n"), pgm_read_word(&rf24_datarate_e_str_P[getDataRate()]));
	printf_P(PSTR("Model\t\t = " PRIPSTR "\r\n"), pgm_read_word(&rf24_model_e_str_P[isPVariant()]));
	printf_P(PSTR("CRC Length\t = " PRIPSTR "\r\n"), pgm_read_word(&rf24_crclength_e_str_P[getCRCLength()]));
	printf_P(PSTR("PA Power\t = " PRIPSTR "\r\n"), pgm_read_word(&rf24_pa_dbm_e_str_P[getPALevel()]));

}

#endif
/****************************************************************************/

bool FuzzyTank::RF24Begin(void)
{
	Serial << F("Initializing RF24.") << endl;
	/*
	FuzzyTank::FuzzyTank(uint16_t _cepin, uint16_t _cspin) :
	ce_pin(_cepin), csn_pin(_cspin), p_variant(false),
	payload_size(32), dynamic_payloads_enabled(false), addr_width(5), csDelay(5)//,pipe0_reading_address(0)
	{
	pipe0_reading_address[0] = 0;
	}
	*/
	//setup the values, without using the initializers
	ce_pin = RF24CePin;
	csn_pin = RF24CsnPin;
	p_variant = false;
	payload_size = 32;
	dynamic_payloads_enabled = false;
	addr_width = 5;
	csDelay = 5;

	uint8_t setup = 0;

#if defined (RF24_LINUX)

#if defined (MRAA)
	GPIO();
	gpio.begin(ce_pin, csn_pin);
#endif

#ifdef RF24_RPi
	switch (csn_pin) {     //Ensure valid hardware CS pin
	case 0: break;
	case 1: break;
		// Allow BCM2835 enums for RPi
	case 8: csn_pin = 0; break;
	case 7: csn_pin = 1; break;
	default: csn_pin = 0; break;
	}
#endif

	_SPI.begin(csn_pin);

	pinMode(ce_pin, OUTPUT);
	ce(LOW);

	delay(100);

#elif defined(LITTLEWIRE)
	pinMode(csn_pin, OUTPUT);
	_SPI.begin();
	csn(HIGH);
#elif defined(XMEGA_D3)
	if (ce_pin != csn_pin) pinMode(ce_pin, OUTPUT);
	_SPI.begin(csn_pin);
	ce(LOW);
	csn(HIGH);
	delay(200);
#else
	// Initialize pins
	if (ce_pin != csn_pin) pinMode(ce_pin, OUTPUT);

#if ! defined(LITTLEWIRE)
	if (ce_pin != csn_pin)
#endif
		pinMode(csn_pin, OUTPUT);

	_SPI.begin();
	ce(LOW);
	csn(HIGH);
#if defined (__ARDUINO_X86__)
	delay(100);
#endif
#endif //Linux

	// Must allow the radio time to settle else configuration bits will not necessarily stick.
	// This is actually only required following power up but some settling time also appears to
	// be required after resets too. For full coverage, we'll always assume the worst.
	// Enabling 16b CRC is by far the most obvious case if the wrong timing is used - or skipped.
	// Technically we require 4.5ms + 14us as a worst case. We'll just call it 5ms for good measure.
	// WARNING: Delay is based on P-variant whereby non-P *may* require different timing.
	delay(5);

	// Reset NRF_CONFIG and enable 16-bit CRC.
	write_register(NRF_CONFIG, 0x0C);

	// Set 1500uS (minimum for 32B payload in ESB@250KBPS) timeouts, to make testing a little easier
	// WARNING: If this is ever lowered, either 250KBS mode with AA is broken or maximum packet
	// sizes must never be used. See documentation for a more complete explanation.
	setRetries(5, 15);

	// Reset value is MAX
	//setPALevel( RF24_PA_MAX ) ;

	// check for connected module and if this is a p nRF24l01 variant
	//
	if (setDataRate(RF24_250KBPS))
	{
		p_variant = true;
	}
	setup = read_register(RF_SETUP);
	/*if( setup == 0b00001110 )     // register default for nRF24L01P
	{
	p_variant = true ;
	}*/

	// Then set the data rate to the slowest (and most reliable) speed supported by all
	// hardware.
	setDataRate(RF24_1MBPS);

	// Initialize CRC and request 2-byte (16bit) CRC
	//setCRCLength( RF24_CRC_16 ) ;

	// Disable dynamic payloads, to match dynamic_payloads_enabled setting - Reset value is 0
	toggle_features();
	write_register(FEATURE, 0);
	write_register(DYNPD, 0);
	dynamic_payloads_enabled = false;

	// Reset current status
	// Notice reset and flush is the last thing we do
	write_register(NRF_STATUS, _BV(RX_DR) | _BV(TX_DS) | _BV(MAX_RT));

	// Set up default configuration.  Callers can always change it later.
	// This channel should be universally safe and not bleed over into adjacent
	// spectrum.
	setChannel(76);

	// Flush buffers
	flush_rx();
	flush_tx();

	powerUp(); //Power up by default when begin() is called

			   // Enable PTX, do not write CE high so radio will remain in standby I mode ( 130us max to transition to RX or TX instead of 1500us from powerUp )
			   // PTX should use only 22uA of power
	write_register(NRF_CONFIG, (read_register(NRF_CONFIG)) & ~_BV(PRIM_RX));

	// if setup is 0 or ff then there was no response from module
	return (setup != 0 && setup != 0xff);
}

/****************************************************************************/

bool FuzzyTank::isChipConnected()
{
	uint8_t setup = read_register(SETUP_AW);
	if (setup >= 1 && setup <= 3)
	{
		return true;
	}

	return false;
}

/****************************************************************************/

void FuzzyTank::startListening(void)
{
#if !defined (RF24_TINY) && ! defined(LITTLEWIRE)
	powerUp();
#endif
	write_register(NRF_CONFIG, read_register(NRF_CONFIG) | _BV(PRIM_RX));
	write_register(NRF_STATUS, _BV(RX_DR) | _BV(TX_DS) | _BV(MAX_RT));
	ce(HIGH);
	// Restore the pipe0 adddress, if exists
	if (pipe0_reading_address[0] > 0) {
		write_register(RX_ADDR_P0, pipe0_reading_address, addr_width);
	}
	else {
		closeReadingPipe(0);
	}

	// Flush buffers
	//flush_rx();
	if (read_register(FEATURE) & _BV(EN_ACK_PAY)) {
		flush_tx();
	}

	// Go!
	//delayMicroseconds(100);
}

/****************************************************************************/
static const uint8_t child_pipe_enable[] PROGMEM =
{
	ERX_P0, ERX_P1, ERX_P2, ERX_P3, ERX_P4, ERX_P5
};

void FuzzyTank::stopListening(void)
{
	ce(LOW);
	delayMicroseconds(txDelay);

	if (read_register(FEATURE) & _BV(EN_ACK_PAY)) {
		delayMicroseconds(txDelay); //200
		flush_tx();
	}
	//flush_rx();

	write_register(NRF_CONFIG, (read_register(NRF_CONFIG)) & ~_BV(PRIM_RX));

#if defined (RF24_TINY) || defined (LITTLEWIRE)
	// for 3 pins solution TX mode is only left with additonal powerDown/powerUp cycle
	if (ce_pin == csn_pin) {
		powerDown();
		powerUp();
	}
#endif

	write_register(EN_RXADDR, read_register(EN_RXADDR) | _BV(pgm_read_byte(&child_pipe_enable[0]))); // Enable RX on pipe0

																									 //delayMicroseconds(100);

}

/****************************************************************************/

void FuzzyTank::powerDown(void)
{
	ce(LOW); // Guarantee CE is low on powerDown
	write_register(NRF_CONFIG, read_register(NRF_CONFIG) & ~_BV(PWR_UP));
}

/****************************************************************************/

//Power up now. Radio will not power down unless instructed by MCU for config changes etc.
void FuzzyTank::powerUp(void)
{
	uint8_t cfg = read_register(NRF_CONFIG);

	// if not powered up then power up and wait for the radio to initialize
	if (!(cfg & _BV(PWR_UP))) {
		write_register(NRF_CONFIG, cfg | _BV(PWR_UP));

		// For nRF24L01+ to go from power down mode to TX or RX mode it must first pass through stand-by mode.
		// There must be a delay of Tpd2stby (see Table 16.) after the nRF24L01+ leaves power down mode before
		// the CEis set high. - Tpd2stby can be up to 5ms per the 1.0 datasheet
		delay(5);
	}
}

/******************************************************************/
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
void FuzzyTank::errNotify() {
#if defined (SERIAL_DEBUG) || defined (RF24_LINUX)
	printf_P(PSTR("RF24 HARDWARE FAIL: Radio not responding, verify pin connections, wiring, etc.\r\n"));
#endif
#if defined (FAILURE_HANDLING)
	failureDetected = 1;
#else
	delay(5000);
#endif
}
#endif
/******************************************************************/

//Similar to the previous write, clears the interrupt flags
bool FuzzyTank::write(const void* buf, uint8_t len, const bool multicast)
{
	//Start Writing
	startFastWrite(buf, len, multicast);

	//Wait until complete or failed
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
	uint32_t timer = millis();
#endif 

	while (!(get_status()  & (_BV(TX_DS) | _BV(MAX_RT)))) {
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
		if (millis() - timer > 95) {
			errNotify();
#if defined (FAILURE_HANDLING)
			return 0;
#else
			delay(100);
#endif
		}
#endif
	}

	ce(LOW);

	uint8_t status = write_register(NRF_STATUS, _BV(RX_DR) | _BV(TX_DS) | _BV(MAX_RT));

	//Max retries exceeded
	if (status & _BV(MAX_RT)) {
		flush_tx(); //Only going to be 1 packet int the FIFO at a time using this method, so just flush
		return 0;
	}
	//TX OK 1 or 0
	return 1;
}

bool FuzzyTank::write(const void* buf, uint8_t len) {
	return write(buf, len, 0);
}
/****************************************************************************/

//For general use, the interrupt flags are not important to clear
bool FuzzyTank::writeBlocking(const void* buf, uint8_t len, uint32_t timeout)
{
	//Block until the FIFO is NOT full.
	//Keep track of the MAX retries and set auto-retry if seeing failures
	//This way the FIFO will fill up and allow blocking until packets go through
	//The radio will auto-clear everything in the FIFO as long as CE remains high

	uint32_t timer = millis();							  //Get the time that the payload transmission started

	while ((get_status()  & (_BV(TX_FULL)))) {		  //Blocking only if FIFO is full. This will loop and block until TX is successful or timeout

		if (get_status() & _BV(MAX_RT)) {					  //If MAX Retries have been reached
			reUseTX();										  //Set re-transmit and clear the MAX_RT interrupt flag
			if (millis() - timer > timeout) { return 0; }		  //If this payload has exceeded the user-defined timeout, exit and return 0
		}
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
		if (millis() - timer > (timeout + 95)) {
			errNotify();
#if defined (FAILURE_HANDLING)
			return 0;
#endif				
		}
#endif

	}

	//Start Writing
	startFastWrite(buf, len, 0);								  //Write the payload if a buffer is clear

	return 1;												  //Return 1 to indicate successful transmission
}

/****************************************************************************/

void FuzzyTank::reUseTX() {
	write_register(NRF_STATUS, _BV(MAX_RT));			  //Clear max retry flag
	spiTrans(REUSE_TX_PL);
	ce(LOW);										  //Re-Transfer packet
	ce(HIGH);
}

/****************************************************************************/

bool FuzzyTank::writeFast(const void* buf, uint8_t len, const bool multicast)
{
	//Block until the FIFO is NOT full.
	//Keep track of the MAX retries and set auto-retry if seeing failures
	//Return 0 so the user can control the retrys and set a timer or failure counter if required
	//The radio will auto-clear everything in the FIFO as long as CE remains high

#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
	uint32_t timer = millis();
#endif

	while ((get_status()  & (_BV(TX_FULL)))) {			  //Blocking only if FIFO is full. This will loop and block until TX is successful or fail

		if (get_status() & _BV(MAX_RT)) {
			//reUseTX();										  //Set re-transmit
			write_register(NRF_STATUS, _BV(MAX_RT));			  //Clear max retry flag
			return 0;										  //Return 0. The previous payload has been retransmitted
															  //From the user perspective, if you get a 0, just keep trying to send the same payload
		}
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
		if (millis() - timer > 95) {
			errNotify();
#if defined (FAILURE_HANDLING)
			return 0;
#endif
		}
#endif
	}
	//Start Writing
	startFastWrite(buf, len, multicast);

	return 1;
}

bool FuzzyTank::writeFast(const void* buf, uint8_t len) {
	return writeFast(buf, len, 0);
}

/****************************************************************************/

//Per the documentation, we want to set PTX Mode when not listening. Then all we do is write data and set CE high
//In this mode, if we can keep the FIFO buffers loaded, packets will transmit immediately (no 130us delay)
//Otherwise we enter Standby-II mode, which is still faster than standby mode
//Also, we remove the need to keep writing the config register over and over and delaying for 150 us each time if sending a stream of data

void FuzzyTank::startFastWrite(const void* buf, uint8_t len, const bool multicast, bool startTx) { //TMRh20

																									   //write_payload( buf,len);
	write_payload(buf, len, multicast ? W_TX_PAYLOAD_NO_ACK : W_TX_PAYLOAD);
	if (startTx) {
		ce(HIGH);
	}

}

/****************************************************************************/

//Added the original startWrite back in so users can still use interrupts, ack payloads, etc
//Allows the library to pass all tests
void FuzzyTank::startWrite(const void* buf, uint8_t len, const bool multicast) {

	// Send the payload

	//write_payload( buf, len );
	write_payload(buf, len, multicast ? W_TX_PAYLOAD_NO_ACK : W_TX_PAYLOAD);
	ce(HIGH);
#if defined(CORE_TEENSY) || !defined(ARDUINO) || defined (RF24_SPIDEV) || defined (RF24_DUE)
	delayMicroseconds(10);
#endif
	ce(LOW);


}

/****************************************************************************/

bool FuzzyTank::rxFifoFull() {
	return read_register(FIFO_STATUS) & _BV(RX_FULL);
}
/****************************************************************************/

bool FuzzyTank::txStandBy() {

#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
	uint32_t timeout = millis();
#endif
	while (!(read_register(FIFO_STATUS) & _BV(TX_EMPTY))) {
		if (get_status() & _BV(MAX_RT)) {
			write_register(NRF_STATUS, _BV(MAX_RT));
			ce(LOW);
			flush_tx();    //Non blocking, flush the data
			return 0;
		}
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
		if (millis() - timeout > 95) {
			errNotify();
#if defined (FAILURE_HANDLING)
			return 0;
#endif
		}
#endif
	}

	ce(LOW);			   //Set STANDBY-I mode
	return 1;
}

/****************************************************************************/

bool FuzzyTank::txStandBy(uint32_t timeout, bool startTx) {

	if (startTx) {
		stopListening();
		ce(HIGH);
	}
	uint32_t start = millis();

	while (!(read_register(FIFO_STATUS) & _BV(TX_EMPTY))) {
		if (get_status() & _BV(MAX_RT)) {
			write_register(NRF_STATUS, _BV(MAX_RT));
			ce(LOW);										  //Set re-transmit
			ce(HIGH);
			if (millis() - start >= timeout) {
				ce(LOW); flush_tx(); return 0;
			}
		}
#if defined (FAILURE_HANDLING) || defined (RF24_LINUX)
		if (millis() - start > (timeout + 95)) {
			errNotify();
#if defined (FAILURE_HANDLING)
			return 0;
#endif
		}
#endif
	}


	ce(LOW);				   //Set STANDBY-I mode
	return 1;

}

/****************************************************************************/

void FuzzyTank::maskIRQ(bool tx, bool fail, bool rx) {

	uint8_t config = read_register(NRF_CONFIG);
	/* clear the interrupt flags */
	config &= ~(1 << MASK_MAX_RT | 1 << MASK_TX_DS | 1 << MASK_RX_DR);
	/* set the specified interrupt flags */
	config |= fail << MASK_MAX_RT | tx << MASK_TX_DS | rx << MASK_RX_DR;
	write_register(NRF_CONFIG, config);
}

/****************************************************************************/

uint8_t FuzzyTank::getDynamicPayloadSize(void)
{
	uint8_t result = 0;

#if defined (RF24_LINUX)  
	spi_txbuff[0] = R_RX_PL_WID;
	spi_rxbuff[1] = 0xff;
	beginTransaction();
	_SPI.transfernb((char *)spi_txbuff, (char *)spi_rxbuff, 2);
	result = spi_rxbuff[1];
	endTransaction();
#else
	beginTransaction();
	_SPI.transfer(R_RX_PL_WID);
	result = _SPI.transfer(0xff);
	endTransaction();
#endif

	if (result > 32) { flush_rx(); delay(2); return 0; }
	return result;
}

/****************************************************************************/

bool FuzzyTank::available(void)
{
	return available(NULL);
}

/****************************************************************************/

bool FuzzyTank::available(uint8_t* pipe_num)
{
	if (!(read_register(FIFO_STATUS) & _BV(RX_EMPTY)))
	{

		// If the caller wants the pipe number, include that
		if (pipe_num) {
			uint8_t status = get_status();
			*pipe_num = (status >> RX_P_NO) & 0x07;
		}
		return 1;
	}


	return 0;


}

/****************************************************************************/

void FuzzyTank::read(void* buf, uint8_t len) {

	// Fetch the payload
	read_payload(buf, len);

	//Clear the two possible interrupt flags with one command
	write_register(NRF_STATUS, _BV(RX_DR) | _BV(MAX_RT) | _BV(TX_DS));

}

/****************************************************************************/

void FuzzyTank::whatHappened(bool& tx_ok, bool& tx_fail, bool& rx_ready)
{
	// Read the status & reset the status in one easy call
	// Or is that such a good idea?
	uint8_t status = write_register(NRF_STATUS, _BV(RX_DR) | _BV(TX_DS) | _BV(MAX_RT));

	// Report to the user what happened
	tx_ok = status & _BV(TX_DS);
	tx_fail = status & _BV(MAX_RT);
	rx_ready = status & _BV(RX_DR);
}

/****************************************************************************/

void FuzzyTank::openWritingPipe(uint64_t value)
{
	// Note that AVR 8-bit uC's store this LSB first, and the NRF24L01(+)
	// expects it LSB first too, so we're good.

	write_register(RX_ADDR_P0, reinterpret_cast<uint8_t*>(&value), addr_width);
	write_register(TX_ADDR, reinterpret_cast<uint8_t*>(&value), addr_width);


	//const uint8_t max_payload_size = 32;
	//write_register(RX_PW_P0,rf24_min(payload_size,max_payload_size));
	write_register(RX_PW_P0, payload_size);
}

/****************************************************************************/
void FuzzyTank::openWritingPipe(const uint8_t *address)
{
	// Note that AVR 8-bit uC's store this LSB first, and the NRF24L01(+)
	// expects it LSB first too, so we're good.

	write_register(RX_ADDR_P0, address, addr_width);
	write_register(TX_ADDR, address, addr_width);

	//const uint8_t max_payload_size = 32;
	//write_register(RX_PW_P0,rf24_min(payload_size,max_payload_size));
	write_register(RX_PW_P0, payload_size);
}

/****************************************************************************/
static const uint8_t child_pipe[] PROGMEM =
{
	RX_ADDR_P0, RX_ADDR_P1, RX_ADDR_P2, RX_ADDR_P3, RX_ADDR_P4, RX_ADDR_P5
};
static const uint8_t child_payload_size[] PROGMEM =
{
	RX_PW_P0, RX_PW_P1, RX_PW_P2, RX_PW_P3, RX_PW_P4, RX_PW_P5
};


void FuzzyTank::openReadingPipe(uint8_t child, uint64_t address)
{
	// If this is pipe 0, cache the address.  This is needed because
	// openWritingPipe() will overwrite the pipe 0 address, so
	// startListening() will have to restore it.
	if (child == 0) {
		memcpy(pipe0_reading_address, &address, addr_width);
	}

	if (child <= 6)
	{
		// For pipes 2-5, only write the LSB
		if (child < 2)
			write_register(pgm_read_byte(&child_pipe[child]), reinterpret_cast<const uint8_t*>(&address), addr_width);
		else
			write_register(pgm_read_byte(&child_pipe[child]), reinterpret_cast<const uint8_t*>(&address), 1);

		write_register(pgm_read_byte(&child_payload_size[child]), payload_size);

		// Note it would be more efficient to set all of the bits for all open
		// pipes at once.  However, I thought it would make the calling code
		// more simple to do it this way.
		write_register(EN_RXADDR, read_register(EN_RXADDR) | _BV(pgm_read_byte(&child_pipe_enable[child])));
	}
}

/****************************************************************************/
void FuzzyTank::setAddressWidth(uint8_t a_width) {

	if (a_width -= 2) {
		write_register(SETUP_AW, a_width % 4);
		addr_width = (a_width % 4) + 2;
	}
	else {
		write_register(SETUP_AW, 0);
		addr_width = 2;
	}

}

/****************************************************************************/

void FuzzyTank::openReadingPipe(uint8_t child, const uint8_t *address)
{
	// If this is pipe 0, cache the address.  This is needed because
	// openWritingPipe() will overwrite the pipe 0 address, so
	// startListening() will have to restore it.
	if (child == 0) {
		memcpy(pipe0_reading_address, address, addr_width);
	}
	if (child <= 6)
	{
		// For pipes 2-5, only write the LSB
		if (child < 2) {
			write_register(pgm_read_byte(&child_pipe[child]), address, addr_width);
		}
		else {
			write_register(pgm_read_byte(&child_pipe[child]), address, 1);
		}
		write_register(pgm_read_byte(&child_payload_size[child]), payload_size);

		// Note it would be more efficient to set all of the bits for all open
		// pipes at once.  However, I thought it would make the calling code
		// more simple to do it this way.
		write_register(EN_RXADDR, read_register(EN_RXADDR) | _BV(pgm_read_byte(&child_pipe_enable[child])));

	}
}

/****************************************************************************/

void FuzzyTank::closeReadingPipe(uint8_t pipe)
{
	write_register(EN_RXADDR, read_register(EN_RXADDR) & ~_BV(pgm_read_byte(&child_pipe_enable[pipe])));
}

/****************************************************************************/

void FuzzyTank::toggle_features(void)
{
	beginTransaction();
	_SPI.transfer(ACTIVATE);
	_SPI.transfer(0x73);
	endTransaction();
}

/****************************************************************************/

void FuzzyTank::enableDynamicPayloads(void)
{
	// Enable dynamic payload throughout the system

	//toggle_features();
	write_register(FEATURE, read_register(FEATURE) | _BV(EN_DPL));


	IF_SERIAL_DEBUG(printf("FEATURE=%i\r\n", read_register(FEATURE)));

	// Enable dynamic payload on all pipes
	//
	// Not sure the use case of only having dynamic payload on certain
	// pipes, so the library does not support it.
	write_register(DYNPD, read_register(DYNPD) | _BV(DPL_P5) | _BV(DPL_P4) | _BV(DPL_P3) | _BV(DPL_P2) | _BV(DPL_P1) | _BV(DPL_P0));

	dynamic_payloads_enabled = true;
}

/****************************************************************************/
void FuzzyTank::disableDynamicPayloads(void)
{
	// Disables dynamic payload throughout the system.  Also disables Ack Payloads

	//toggle_features();
	write_register(FEATURE, 0);


	IF_SERIAL_DEBUG(printf("FEATURE=%i\r\n", read_register(FEATURE)));

	// Disable dynamic payload on all pipes
	//
	// Not sure the use case of only having dynamic payload on certain
	// pipes, so the library does not support it.
	write_register(DYNPD, 0);

	dynamic_payloads_enabled = false;
}

/****************************************************************************/

void FuzzyTank::enableAckPayload(void)
{
	//
	// enable ack payload and dynamic payload features
	//

	//toggle_features();
	write_register(FEATURE, read_register(FEATURE) | _BV(EN_ACK_PAY) | _BV(EN_DPL));

	IF_SERIAL_DEBUG(printf("FEATURE=%i\r\n", read_register(FEATURE)));

	//
	// Enable dynamic payload on pipes 0 & 1
	//

	write_register(DYNPD, read_register(DYNPD) | _BV(DPL_P1) | _BV(DPL_P0));
	dynamic_payloads_enabled = true;
}

/****************************************************************************/

void FuzzyTank::enableDynamicAck(void) {
	//
	// enable dynamic ack features
	//
	//toggle_features();
	write_register(FEATURE, read_register(FEATURE) | _BV(EN_DYN_ACK));

	IF_SERIAL_DEBUG(printf("FEATURE=%i\r\n", read_register(FEATURE)));


}

/****************************************************************************/

void FuzzyTank::writeAckPayload(uint8_t pipe, const void* buf, uint8_t len)
{
	const uint8_t* current = reinterpret_cast<const uint8_t*>(buf);

	uint8_t data_len = rf24_min(len, 32);

#if defined (RF24_LINUX)
	beginTransaction();
	uint8_t * ptx = spi_txbuff;
	uint8_t size = data_len + 1; // Add register value to transmit buffer
	*ptx++ = W_ACK_PAYLOAD | (pipe & 0x07);
	while (data_len--) {
		*ptx++ = *current++;
	}

	_SPI.transfern((char *)spi_txbuff, size);
	endTransaction();
#else
	beginTransaction();
	_SPI.transfer(W_ACK_PAYLOAD | (pipe & 0x07));

	while (data_len--)
		_SPI.transfer(*current++);
	endTransaction();

#endif  

}

/****************************************************************************/

bool FuzzyTank::isAckPayloadAvailable(void)
{
	return !(read_register(FIFO_STATUS) & _BV(RX_EMPTY));
}

/****************************************************************************/

bool FuzzyTank::isPVariant(void)
{
	return p_variant;
}

/****************************************************************************/

void FuzzyTank::setAutoAck(bool enable)
{
	if (enable)
		write_register(EN_AA, 0x3F);
	else
		write_register(EN_AA, 0);
}

/****************************************************************************/

void FuzzyTank::setAutoAck(uint8_t pipe, bool enable)
{
	if (pipe <= 6)
	{
		uint8_t en_aa = read_register(EN_AA);
		if (enable)
		{
			en_aa |= _BV(pipe);
		}
		else
		{
			en_aa &= ~_BV(pipe);
		}
		write_register(EN_AA, en_aa);
	}
}

/****************************************************************************/

bool FuzzyTank::testCarrier(void)
{
	return (read_register(CD) & 1);
}

/****************************************************************************/

bool FuzzyTank::testRPD(void)
{
	return (read_register(RPD) & 1);
}

/****************************************************************************/

void FuzzyTank::setPALevel(uint8_t level)
{

	uint8_t setup = read_register(RF_SETUP) & 0xF8;

	if (level > 3) {  						// If invalid level, go to max PA
		level = (RF24_PA_MAX << 1) + 1;		// +1 to support the SI24R1 chip extra bit
	}
	else {
		level = (level << 1) + 1;	 		// Else set level as requested
	}


	write_register(RF_SETUP, setup |= level);	// Write it to the chip
}

/****************************************************************************/

uint8_t FuzzyTank::getPALevel(void)
{

	return (read_register(RF_SETUP) & (_BV(RF_PWR_LOW) | _BV(RF_PWR_HIGH))) >> 1;
}

/****************************************************************************/

bool FuzzyTank::setDataRate(rf24_datarate_e speed)
{
	bool result = false;
	uint8_t setup = read_register(RF_SETUP);

	// HIGH and LOW '00' is 1Mbs - our default
	setup &= ~(_BV(RF_DR_LOW) | _BV(RF_DR_HIGH));

#if defined(__arm__) || defined (RF24_LINUX) || defined (__ARDUINO_X86__)
	txDelay = 250;
#else //16Mhz Arduino
	txDelay = 85;
#endif
	if (speed == RF24_250KBPS)
	{
		// Must set the RF_DR_LOW to 1; RF_DR_HIGH (used to be RF_DR) is already 0
		// Making it '10'.
		setup |= _BV(RF_DR_LOW);
#if defined(__arm__) || defined (RF24_LINUX) || defined (__ARDUINO_X86__)
		txDelay = 450;
#else //16Mhz Arduino
		txDelay = 155;
#endif
	}
	else
	{
		// Set 2Mbs, RF_DR (RF_DR_HIGH) is set 1
		// Making it '01'
		if (speed == RF24_2MBPS)
		{
			setup |= _BV(RF_DR_HIGH);
#if defined(__arm__) || defined (RF24_LINUX) || defined (__ARDUINO_X86__)
			txDelay = 190;
#else //16Mhz Arduino	  
			txDelay = 65;
#endif
		}
	}
	write_register(RF_SETUP, setup);

	// Verify our result
	if (read_register(RF_SETUP) == setup)
	{
		result = true;
	}
	return result;
}

/****************************************************************************/

rf24_datarate_e FuzzyTank::getDataRate(void)
{
	rf24_datarate_e result;
	uint8_t dr = read_register(RF_SETUP) & (_BV(RF_DR_LOW) | _BV(RF_DR_HIGH));

	// switch uses RAM (evil!)
	// Order matters in our case below
	if (dr == _BV(RF_DR_LOW))
	{
		// '10' = 250KBPS
		result = RF24_250KBPS;
	}
	else if (dr == _BV(RF_DR_HIGH))
	{
		// '01' = 2MBPS
		result = RF24_2MBPS;
	}
	else
	{
		// '00' = 1MBPS
		result = RF24_1MBPS;
	}
	return result;
}

/****************************************************************************/

void FuzzyTank::setCRCLength(rf24_crclength_e length)
{
	uint8_t config = read_register(NRF_CONFIG) & ~(_BV(CRCO) | _BV(EN_CRC));

	// switch uses RAM (evil!)
	if (length == RF24_CRC_DISABLED)
	{
		// Do nothing, we turned it off above.
	}
	else if (length == RF24_CRC_8)
	{
		config |= _BV(EN_CRC);
	}
	else
	{
		config |= _BV(EN_CRC);
		config |= _BV(CRCO);
	}
	write_register(NRF_CONFIG, config);
}

/****************************************************************************/

rf24_crclength_e FuzzyTank::getCRCLength(void)
{
	rf24_crclength_e result = RF24_CRC_DISABLED;

	uint8_t config = read_register(NRF_CONFIG) & (_BV(CRCO) | _BV(EN_CRC));
	uint8_t AA = read_register(EN_AA);

	if (config & _BV(EN_CRC) || AA)
	{
		if (config & _BV(CRCO))
			result = RF24_CRC_16;
		else
			result = RF24_CRC_8;
	}

	return result;
}

/****************************************************************************/

void FuzzyTank::disableCRC(void)
{
	uint8_t disable = read_register(NRF_CONFIG) & ~_BV(EN_CRC);
	write_register(NRF_CONFIG, disable);
}

/****************************************************************************/
void FuzzyTank::setRetries(uint8_t delay, uint8_t count)
{
	write_register(SETUP_RETR, (delay & 0xf) << ARD | (count & 0xf) << ARC);
}


//ATTiny support code pulled in from https://github.com/jscrane/RF24

#if defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__) || defined(__AVR_ATtiny85__)
// see http://gammon.com.au/spi
#	define DI   0  // D0, pin 5  Data In
#	define DO   1  // D1, pin 6  Data Out (this is *not* MOSI)
#	define USCK 2  // D2, pin 7  Universal Serial Interface clock
#	define SS   3  // D3, pin 2  Slave Select
#elif defined(__AVR_ATtiny24__) || defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
// these depend on the core used (check pins_arduino.h)
// this is for jeelabs' one (based on google-code core)
#	define DI   4   // PA6
#	define DO   5   // PA5
#	define USCK 6   // PA4
#	define SS   3   // PA7
#elif defined(__AVR_ATtiny2313__) || defined(__AVR_ATtiny4313__)
// these depend on the core used (check pins_arduino.h)
// tested with google-code core
#	define DI   14  // PB5
#	define DO   15  // PB6
#	define USCK 16  // PB7
#	define SS   13  // PB4
#elif defined(__AVR_ATtiny861__)
// these depend on the core used (check pins_arduino.h)
// tested with google-code core
#    define DI   9   // PB0
#    define DO   8   // PB1
#    define USCK 7   // PB2
#    define SS   6   // PB3
#endif

#if defined(RF24_TINY)

void SPIClass::begin() {

	pinMode(USCK, OUTPUT);
	pinMode(DO, OUTPUT);
	pinMode(DI, INPUT);
	USICR = _BV(USIWM0);

}

byte SPIClass::transfer(byte b) {

	USIDR = b;
	USISR = _BV(USIOIF);
	do
		USICR = _BV(USIWM0) | _BV(USICS1) | _BV(USICLK) | _BV(USITC);
	while ((USISR & _BV(USIOIF)) == 0);
	return USIDR;

}

void SPIClass::end() {}
void SPIClass::setDataMode(uint8_t mode) {}
void SPIClass::setBitOrder(uint8_t bitOrder) {}
void SPIClass::setClockDivider(uint8_t rate) {}


#endif
#pragma endregion RF24

#pragma region Custom RF24
void FuzzyTank::updateRF24()
{
	while (available())
	{              

		uint8_t dpl = getDynamicPayloadSize();
		read(&payload, dpl);

		#ifdef DEBUG_OUTPUT_CONTROL_DATA
		Serial << ("rx: ");
		for (uint8_t i = 0; i < dpl; i++)
		{
			Serial << " " << payload[i];
		}
		Serial << endl;
		#endif //DEBUG_OUTPUT_CONTROL_DATA

		
		
		ack = ACK_CMD_V1_OK;
		writeAckPayload(1, &ack, 1);
		signalTimer = millis();
		handleRF24Packet();
	}
	if (millis() - signalTimer > SIGNAL_LOSS_MOTOR_OFF_PERIOD)
	{
		commandMotorDirect(0, 0);
	}
}
void FuzzyTank::handleRF24Packet()
{
	switch (payload[0])
	{
	case CMD_BASIC_CONTROL_V1:
	{
		joystickX = payload[1];
		joystickY = payload[2];
		previousButtonState = buttonState;
		buttonState = payload[3];
		commandTank();
		break;
	}
	default:
		break;
	}
}
bool FuzzyTank::button(uint8_t btn)
{
	return (buttonState & btn);
}
bool FuzzyTank::pressed(uint8_t btn)
{
	return !(previousButtonState & btn) && (buttonState & btn);
}
bool FuzzyTank::released(uint8_t btn)
{
	return (previousButtonState & btn) && !(buttonState & btn);
}
#pragma endregion Custom RF24

#pragma region MOTOR Control
void FuzzyTank::motorBegin()
{
	Serial << F("Initializing motor.") << endl;
	pinMode(motorAPwmPin, OUTPUT);
	digitalWrite(motorAPwmPin, LOW);
	pinMode(motorAIn1Pin, OUTPUT);
	digitalWrite(motorAIn1Pin, LOW);
	pinMode(motorAIn2Pin, OUTPUT);
	digitalWrite(motorAIn2Pin, LOW);
	pinMode(motorBPwmPin, OUTPUT);
	digitalWrite(motorBPwmPin, LOW);
	pinMode(motorBIn1Pin, OUTPUT);
	digitalWrite(motorBIn1Pin, LOW);
	pinMode(motorBIn2Pin, OUTPUT);
	digitalWrite(motorBIn2Pin, LOW);

	motorFilterShift = DEFAULT_MOTOR_FILTER_SHIFT;
	maxForwardSpeed = DEFAULT_MAX_FORWARD_SPEED;
	maxBackwardSpeed = DEFAULT_MAX_BACKWARD_SPEED;
	maxTurnSpeed = DEFAULT_MAX_TURN_SPEED;
	maxSpeed = DEFAULT_DIRECT_MAX_SPEED;
	motorPWMLowThreshold = DEFAULT_MOTOR_PWM_LOW_THRESHOLD;

	//set pwm frequency
	//Setting 	Divisor 	Frequency
	//	0x01 	 	1 	 	62500
	//	0x02  		8 	 	7812.5
	//	0x03  		64 	 	976.5625   <--DEFAULT
	//	0x04 	 	256 	 	244.140625
	//	0x05 	 	1024 	 	61.03515625

	//TCCR0B = (TCCR0B & 0b11111000) | 0x02;

	//set motor variations
	switch (motorType)
	{
		case DEFAULT_MOTOR:
		{
			
			break;
		}
		case TB16:
		{
			
			break;
		}
		case GA20:
		{
			
			break;
		}
		default:
		{
			
			break;
		}
	}
	//set gearset variations
	switch (gearsetType)
	{
		case DEFAULT_GEARSET:
		{
			break;
		}
		case TOWER_GEARSET_50_TO_1:
		{
			motorOutputDirection = 1;
			break;
		}
		case SY_PLANETARY_GEARSET_53_TO_1:
		{
			motorOutputDirection = -1;
			break;
		}
		case SY_PLANETARY_GEARSET_29_TO_1:
		{
			motorOutputDirection = -1;
			break;
		}
		default:
		{
			break;
		}
	}
}

void FuzzyTank::motorDebug(uint8_t x, uint8_t y)
{
	joystickX = x;
	joystickY = y;
	commandTank();
}

void FuzzyTank::setMotorSmoothness(uint8_t value)
{
	motorFilterShift = value;
}

void FuzzyTank::setMaximumSpeed(uint8_t speed)
{
	if (speed > motorPWMLowThreshold)
	{
		maxForwardSpeed = speed;
		maxBackwardSpeed = speed;
	}
}

void FuzzyTank::setMaximumForwardSpeed(uint8_t speed)
{
	if (speed > motorPWMLowThreshold)
	{
		maxForwardSpeed = speed;
	}
}

void FuzzyTank::setMaximumBackwardSpeed(uint8_t speed)
{
	if (speed > motorPWMLowThreshold)
	{
		maxBackwardSpeed = speed;
	}
}

void FuzzyTank::setMaximumTurnSpeed(uint8_t speed)
{
	if (speed > motorPWMLowThreshold)
	{
		maxTurnSpeed = speed;
	}
}

#pragma endregion MOTOR Control

#pragma region Communication and Control
void FuzzyTank::commandTank()
{
	/*
	if (pressed(BUTTON_SELECT)) Serial << "Select pressed";
	if (pressed(BUTTON_START)) Serial << "Start pressed";
	if (pressed(BUTTON_B)) Serial << "B pressed";
	if (pressed(BUTTON_A)) Serial << "A pressed";
	if (released(BUTTON_SELECT)) Serial << "Select released";
	if (released(BUTTON_START)) Serial << "Start released";
	if (released(BUTTON_B)) Serial << "B released";
	if (released(BUTTON_A)) Serial << "A released";
	if (button(BUTTON_SELECT)) Serial << "Select on";
	if (button(BUTTON_START)) Serial << "Start on";
	if (button(BUTTON_B)) Serial << "B on";
	if (button(BUTTON_A)) Serial << "A on";
	Serial << endl;
	*/

	int8_t x, y; //joystick value adjusted with low threshold and high threshold, -128~127
	int8_t l, r; //converted to left and right motor movement, -128~127

	#ifdef DEBUG_MOTOR_OUTPUT
	Serial << "JX " << joystickX << " JY " << joystickY;
	#endif //DEBUG_MOTOR_OUTPUT

	if (joystickY > JY_CENTER + JY_LOW_THRESHOLD)
	{
		y = map(joystickY, JY_CENTER + JY_LOW_THRESHOLD, JY_CENTER + JY_HIGH_THRESHOLD, 1, 127);
		if (joystickY > JY_CENTER + JY_HIGH_THRESHOLD)
		{
			y = 127;
		}
	}
	else if (joystickY < JY_CENTER - JY_LOW_THRESHOLD)
	{
		y = map(joystickY, JY_CENTER - JY_LOW_THRESHOLD, JY_CENTER - JY_HIGH_THRESHOLD, -1, -128);
		if (joystickY < JY_CENTER - JY_HIGH_THRESHOLD)
		{
			y = -128;
		}
	}
	else
	{
		y = 0;
	}

	if (joystickX > JX_CENTER + JX_LOW_THRESHOLD)
	{
		x = map(joystickX, JX_CENTER + JX_LOW_THRESHOLD, JX_CENTER + JX_HIGH_THRESHOLD, 1, 127);
		if (joystickX > JX_CENTER + JX_HIGH_THRESHOLD)
		{
			x = 127;
		}
	}
	else if (joystickX < JX_CENTER - JX_LOW_THRESHOLD)
	{
		x = map(joystickX, JX_CENTER - JX_LOW_THRESHOLD, JX_CENTER - JX_HIGH_THRESHOLD, -1, -128);
		if (joystickX < JX_CENTER - JX_HIGH_THRESHOLD)
		{
			x = -128;
		}
	}
	else
	{
		x = 0;
	}

	if (y > 0) //forward
	{
		if (x > 0) //forward right
		{
			l = y;
			r = map(x, 1, 127, y, 1);
		}
		else if (x < 0) //forward left
		{
			l = map(x, -1, -128, y, 1);
			r = y;
		}
		else //forward straight
		{
			l = y;
			r = y;
		}
	}
	else if (y < 0) //backward
	{
		if (x > 0) //backward right
		{
			l = y;
			r = map(x, 1, 127, y, -1);
		}
		else if (x < 0) //backward left
		{
			l = map(x, -1, -128, y, -1);
			r = y;
		}
		else //forward straight
		{
			l = y;
			r = y;
		}
	}
	else //(y=0) 
	{
		if (x > 0) //right rotation
		{
			l = map(x, 1, 127, 1, maxTurnSpeed);
			r = map(x, 1, 127, -1, -maxTurnSpeed);
		}
		else if (x < 0) //left rotation
		{
			l = map(x, -1, -128, -1, -maxTurnSpeed);
			r = map(x, -1, -128, 1, maxTurnSpeed);
		}
		else //still
		{
			l = 0;
			r = 0;
		}

	}

	//Serial << "y = " << y << " x = " << x  << "    l = " << l << " r = " << r << endl;
	handleButtons();
	commandMotor(l, r);
}

void FuzzyTank::commandMotor(int8_t motorLeft, int8_t motorRight)
{
	#ifdef DEBUG_MOTOR_OUTPUT
	Serial << " RL " << motorLeft << " RR " << motorRight;
	#endif //DEBUG_MOTOR_OUTPUT
	
	//perform low pass filtering
	if (motorLeftFilterRegister > 0)
	{
		motorLeftFilterRegister = motorLeftFilterRegister - (motorLeftFilterRegister >> motorFilterShift) + motorLeft;
		motorLeftOutput = motorLeftFilterRegister >> motorFilterShift;
	}
	else
	{
		motorLeftFilterRegister = motorLeftFilterRegister - (-((-motorLeftFilterRegister) >> motorFilterShift)) + motorLeft;
		motorLeftOutput = -((-motorLeftFilterRegister) >> motorFilterShift);
	}
	if (motorRightFilterRegister > 0)
	{
		motorRightFilterRegister = motorRightFilterRegister - (motorRightFilterRegister >> motorFilterShift) + motorRight;
		motorRightOutput = motorRightFilterRegister >> motorFilterShift;
	}
	else
	{
		motorRightFilterRegister = motorRightFilterRegister - (-((-motorRightFilterRegister) >> motorFilterShift)) + motorRight;
		motorRightOutput = -((-motorRightFilterRegister) >> motorFilterShift);
	}
	

	#ifdef DEBUG_MOTOR_OUTPUT
	Serial << " OL " << motorLeftOutput << " OR " << motorRightOutput;
	
	#endif //DEBUG_MOTOR_OUTPUT
	
	//transform to proper scale
	if (motorLeftOutput > 0)
	{
		motorLeftPWM = map(motorLeftOutput, 1, 127, motorPWMLowThreshold, maxForwardSpeed);
	}
	else if (motorLeftOutput < 0)
	{
		motorLeftPWM = map(motorLeftOutput, -1, -128, motorPWMLowThreshold, maxBackwardSpeed);
	}
	else if (motorLeftOutput == 0)
	{
		motorLeftPWM = 0;
	}
	if (motorRightOutput > 0)
	{
		motorRightPWM = map(motorRightOutput, 1, 127, motorPWMLowThreshold, maxForwardSpeed);
	}
	else if (motorRightOutput < 0)
	{
		motorRightPWM = map(motorRightOutput, -1, -128, motorPWMLowThreshold, maxBackwardSpeed);
	}
	else if (motorRightOutput == 0)
	{
		motorRightPWM = 0;
	}

	#ifdef DEBUG_MOTOR_OUTPUT
	Serial << " PL " << motorLeftPWM << " PR " << motorRightPWM;
	Serial << endl;
	#endif //DEBUG_MOTOR_OUTPUT
	

	//command values to h-bridge
	if ((motorLeftOutput * motorOutputDirection) > 0)
	{
		digitalWrite(motorAIn1Pin, HIGH);
		digitalWrite(motorAIn2Pin, LOW);
		analogWrite(motorAPwmPin, motorLeftPWM);
	}
	else if ((motorLeftOutput * motorOutputDirection) < 0)
	{
		digitalWrite(motorAIn1Pin, LOW);
		digitalWrite(motorAIn2Pin, HIGH);
		analogWrite(motorAPwmPin, motorLeftPWM);
	}
	else
	{
		digitalWrite(motorAIn1Pin, LOW);
		digitalWrite(motorAIn2Pin, LOW);
		analogWrite(motorAPwmPin,0);
	}

	if ((motorRightOutput * motorOutputDirection) > 0)
	{
		digitalWrite(motorBIn1Pin, LOW);
		digitalWrite(motorBIn2Pin, HIGH);
		analogWrite(motorBPwmPin, motorRightPWM);
	}
	else if ((motorRightOutput * motorOutputDirection) < 0)
	{
		digitalWrite(motorBIn1Pin, HIGH);
		digitalWrite(motorBIn2Pin, LOW);
		analogWrite(motorBPwmPin, motorRightPWM);
	}
	else
	{
		digitalWrite(motorBIn1Pin, LOW);
		digitalWrite(motorBIn2Pin, LOW);
		analogWrite(motorBPwmPin, 0);
	}
	

	/*
	if (motorLeft > 0)
	{
		digitalWrite(MOTOR_A_IN1_PIN, HIGH);
		digitalWrite(MOTOR_A_IN2_PIN, LOW);
		analogWrite(MOTOR_A_PWM_PIN, map(motorLeft, 1, 127, MOTOR_PWM_LOW_THRESHOLD, maxForwardSpeed));
	}
	else if (motorLeft < 0)
	{
		digitalWrite(MOTOR_A_IN1_PIN, LOW);
		digitalWrite(MOTOR_A_IN2_PIN, HIGH);
		analogWrite(MOTOR_A_PWM_PIN, map(motorLeft, -1, -128, MOTOR_PWM_LOW_THRESHOLD, maxBackwardSpeed));
	}
	else
	{
		digitalWrite(MOTOR_A_IN1_PIN, LOW);
		digitalWrite(MOTOR_A_IN2_PIN, LOW);
		analogWrite(MOTOR_A_PWM_PIN, 0);
	}

	if (motorRight > 0)
	{
		digitalWrite(MOTOR_B_IN1_PIN, LOW);
		digitalWrite(MOTOR_B_IN2_PIN, HIGH);
		analogWrite(MOTOR_B_PWM_PIN, map(motorRight, 1, 127, MOTOR_PWM_LOW_THRESHOLD, maxForwardSpeed));
	}
	else if (motorRight < 0)
	{
		digitalWrite(MOTOR_B_IN1_PIN, HIGH);
		digitalWrite(MOTOR_B_IN2_PIN, LOW);
		analogWrite(MOTOR_B_PWM_PIN, map(motorRight, -1, -128, MOTOR_PWM_LOW_THRESHOLD, maxBackwardSpeed));
	}
	else
	{
		digitalWrite(MOTOR_B_IN1_PIN, LOW);
		digitalWrite(MOTOR_B_IN2_PIN, LOW);
		analogWrite(MOTOR_B_PWM_PIN, 0);
	}
	*/
}

void FuzzyTank::commandMotorDirect(int8_t motorLeft, int8_t motorRight)
{
	int8_t motorLeftOutput = motorLeft* motorOutputDirection;
	int8_t motorRightOutput = motorRight* motorOutputDirection;

	if (motorLeftOutput > 0)
	{
		digitalWrite(motorAIn1Pin, HIGH);
		digitalWrite(motorAIn2Pin, LOW);
		analogWrite(motorAPwmPin, map(motorLeftOutput, 1, 127, motorPWMLowThreshold, maxSpeed));
	}
	else if (motorLeftOutput < 0)
	{
		digitalWrite(motorAIn1Pin, LOW);
		digitalWrite(motorAIn2Pin, HIGH);
		analogWrite(motorAPwmPin, map(motorLeftOutput, -1, -128, motorPWMLowThreshold, maxSpeed));
	}
	else
	{
		digitalWrite(motorAIn1Pin, LOW);
		digitalWrite(motorAIn2Pin, LOW);
		analogWrite(motorAPwmPin, 0);
	}

	if (motorRightOutput > 0)
	{
		digitalWrite(motorBIn1Pin, LOW);
		digitalWrite(motorBIn2Pin, HIGH);
		analogWrite(motorBPwmPin, map(motorRightOutput, 1, 127, motorPWMLowThreshold, maxSpeed));
	}
	else if (motorRightOutput < 0)
	{
		digitalWrite(motorBIn1Pin, HIGH);
		digitalWrite(motorBIn2Pin, LOW);
		analogWrite(motorBPwmPin, map(motorRightOutput, -1, -128, motorPWMLowThreshold, maxSpeed));
	}
	else
	{
		digitalWrite(motorBIn1Pin, LOW);
		digitalWrite(motorBIn2Pin, LOW);
		analogWrite(motorBPwmPin, 0);
	}
}

void FuzzyTank::handleButtons()
{
	if (button(BUTTON_SELECT) && button(BUTTON_START) && button(BUTTON_B) && pressed(BUTTON_A))
	{
		//four buttons pressed at the same time, with A pressed last, enter auto-pairing mode
		autoPairing();
	}
	else
	{
		if (pressed(BUTTON_A))
		{
			if (isReloading == false) fire();
		}
		if (pressed(BUTTON_B))
		{
			#ifdef DEBUG_GAMEPLAY
			hit();
			#endif
		}
		if (pressed(BUTTON_START))
		{
			#ifdef DEBUG_GAMEPLAY
			life = 1;
			blood = 10;
			hit();
			#endif
		}
		if (pressed(BUTTON_SELECT))
		{
			#ifdef DEBUG_GAMEPLAY
			life = 2;
			blood = 10;
			hit();
			#endif
		}
	}

}

void FuzzyTank::fire()
{
	Serial << F("Firing!") << endl;
	frontLedOn(200);
	playMp3(MP3_SF_FIRING_SOUND);
	commandMotorDirect(-30, -30);
	delay(100);
	commandMotorDirect(15, 15);
	delay(200);
	commandMotorDirect(0, 0);

	//static uint16_t address = 0xF0F1;
	//static uint32_t command = 0xA45B;


	IRLwrite<IR_NEC>(IRSenderPin, irSendPacket.NEC_address.address, irSendPacket.NEC_command.command);
	irSendPacket.NEC_address.byte1++;
	irSendPacket.NEC_address.byte2++;
	irSendPacket.NEC_command.byte3++;
	irSendPacket.NEC_command.commandNot = ~irSendPacket.NEC_command.byte3;

	isReloading = true;
	setFrontLedDimOff();
	reloadingTimer = millis();
}

void FuzzyTank::dead()
{
	life = life - 1;
	Serial << F("Dead! remaining life = ") << life << endl;

	tankState = TANK_STATE_DEAD;

	commandMotorDirect(-60, 60);
	delay(500);
	enableServo1();
	setServo1(SERVO1_HALF_EXTENDED_POSITION);
	commandMotorDirect(-80, 80);
	delay(500);
	for (uint16_t pos = SERVO1_HALF_EXTENDED_POSITION; pos < SERVO1_HALF_EXTENDED_POSITION; pos = pos + 50)
	{
		setServo1(pos);
		delay(30);
	}
	commandMotorDirect(-100, 100);
	for (uint16_t pos = SERVO1_HALF_EXTENDED_POSITION; pos < SERVO1_FULL_EXTENDED_POSITION; pos = pos + 20)
	{
		setServo1(pos);
		delay(30);
	}
	disableServo1();
	commandMotorDirect(0, 0);

	if (life > 0)
	{
		
		revive();
	}
	else
	{
		gameOver();
	}
}

void FuzzyTank::revive()
{
	Serial << F("Revive!") << endl;
	tankState = TANK_STATE_NORMAL;
	
	playMp3(MP3_SF_REVIVE);
	delay(2000);

	enableServo1();
	setServo1(SERVO1_FULL_EXTENDED_POSITION-200);
	commandMotorDirect(reviveMotorCommandValue, reviveMotorCommandValue);
	delay(300);
	commandMotorDirect(reviveMotorCommandValue/2, reviveMotorCommandValue/2);
	
	for (uint16_t pos = SERVO1_FULL_EXTENDED_POSITION-200; pos > SERVO1_RETRACTED_POSITION; pos = pos - 20)
	{
		setServo1(pos);
		delay(30);
	}
	commandMotorDirect(0, 0);
	setServo1AndOff(SERVO1_RETRACTED_POSITION, 1000);

	blood = DEFAULT_BLOOD;
	bloodState = BLOOD_STATE_NORMAL;
	isInvincible = true;
	invincibleTimer = millis();
	flashHeadLed = false;
	setHeadColor(headColor);
	updateColor();
	reloadingPeriod = DEFAULT_RELOADING_PERIOD;
}

void FuzzyTank::gameOver()
{
	if (isGameOver == true)hitAfterGameover();

	Serial << F("Game over!") << endl;
	playMp3(MP3_SF_DEAD);
	setHeadColor(headColor);
	setEyesColor(eyesColor);
	updateColor();

	delay(2000);
	digitalWrite(frontLedPin, LOW);
	commandMotorDirect(0, 0);
	setHeadColor(fixColor[C_OFF]);
	setEyesColor(fixColor[C_OFF]);
	updateColor();
	
	isGameOver = true;
	
	if (canBeHitAfterGameoverRegister == false)
	{
		while (true);
	}
}

void FuzzyTank::hit()
{
	if (isInvincible == true) return;

	if (isGameOver == true) 
	{
		hitAfterGameover();
	}
	else
	{
		blood = blood - NORMAL_SHOT_DAMAGE;

		Serial << F("Got hit! Remaining blood = ") << blood << endl;

		if (blood > 10)
		{
			playMp3(MP3_SF_HIT);
		}
		else
		{
			playMp3(MP3_SF_MAJOR_HIT);
		}

		if (blood > 0)
		{
			if (blood <= DEFAULT_BLOOD_CRITICAL_THRESHOLD)
			{
				flashHeadLed = true;
				bloodState = BLOOD_STATE_CRITICAL;
				bloodStateFlashingTimer = millis();
				bloodStateFlashingDuration = BLOOD_CRITICAL_FLASHING_PERIOD;
			}
			else if (blood <= DEFAULT_BLOOD_LOW_THRESHOLD)
			{
				flashHeadLed = true;
				bloodState = BLOOD_STATE_LOW;
				bloodStateFlashingTimer = millis();
				bloodStateFlashingDuration = BLOOD_LOW_FLASHING_PERIOD;
			}

			
			uint8_t direction;
			if (random(100) > 50)
			{
				direction = 1;
			}
			else
			{
				direction = -1;
			}
			int8_t speed = 30;
			uint16_t time = 200;
			enableServo1();
			commandMotorDirect(0, 0);
			setServo1(SERVO1_HALF_EXTENDED_POSITION);
			delay(time);
			commandMotorDirect(speed * direction, speed * (-direction));
			delay(time);
			commandMotorDirect(speed * (-direction), speed * direction);
			delay(time);
			commandMotorDirect(speed * direction, speed * (-direction));
			delay(time);
			commandMotorDirect(speed * (-direction), speed * direction);
			delay(time);
			setServo1(SERVO1_RETRACTED_POSITION);
			commandMotorDirect(speed * direction, speed * (-direction));
			delay(time);
			commandMotorDirect(speed * (-direction), speed * direction);
			delay(time);
			commandMotorDirect(0, 0);
			delay(time);

			setServo1AndOff(SERVO1_RETRACTED_POSITION, 1000);
			//disableServo1();

			setHeadColor(fixColor[C_OFF]);
			updateColor();
			isInvincible = true;
			if (bloodState == BLOOD_STATE_CRITICAL)
			{
				reloadingPeriod = DEFAULT_RELOADING_PERIOD_BLOOD_CRITICAL;
			}
			else
			{
				reloadingPeriod = DEFAULT_RELOADING_PERIOD_DURING_INVINCIBLE;
			}

			invincibleTimer = millis();
		}
		else
		{
			dead();
		}
	}
}

void FuzzyTank::hitAfterGameover()
{
	
	Serial << F("Hit after gameover!") << endl;
	setHeadColor(headColor);
	setEyesColor(eyesColor);
	updateColor();
	uint8_t randomValue = random(100);
	if (randomValue < 30)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_1);
	}
	else if (randomValue < 40)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_2);
	}
	else if (randomValue < 50)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_3);
	}
	else if (randomValue < 60)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_4);
	}
	else if (randomValue < 70)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_5);
	}
	else if (randomValue < 80)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_6);
	}
	else if (randomValue < 90)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_7);
	}
	else if (randomValue < 100)
	{
		playMp3(MP3_SF_DEAD_HIT_RANDOM_8);
	}
	if (random(100) < 50)
	{
		commandMotorDirect(-50, 50);
	}
	else
	{
		commandMotorDirect(50, -50);
	}
	
	delay(100);
	commandMotorDirect(0, 0);
	setHeadColor(fixColor[C_OFF]);
	setEyesColor(fixColor[C_OFF]);
	updateColor();
}


#pragma endregion Communication and Control

#pragma region Game play
void FuzzyTank::updateGamePlay()
{
	if (isReloading == true)
	{
		if (millis() - reloadingTimer > reloadingPeriod)
		{
			isReloading = false;
			setFrontLedDimOn();
		}
	}
	if (isInvincible == true)
	{
		if (millis() - invincibleTimer > invincibleDuration)
		{
			isInvincible = false;
			setHeadColor(headColor);
			updateColor();
			if (bloodState == BLOOD_STATE_CRITICAL)
			{
				reloadingPeriod = DEFAULT_RELOADING_PERIOD_BLOOD_CRITICAL;
			}
			else
			{
				reloadingPeriod = DEFAULT_RELOADING_PERIOD;
			}
		}
	}
}
void FuzzyTank::setBlood(uint8_t bloodValue)
{
	blood = bloodValue;
}

void FuzzyTank::setLife(uint8_t lifeValue)
{
	life = lifeValue;
}
void FuzzyTank::canBeHitAfterGameover(bool value)
{
	canBeHitAfterGameoverRegister = value;
}
#pragma endregion Game play

#pragma region IR
uint32_t IRL_LastTime = 0;
uint8_t IRProtocol;
uint16_t IRAddress;
uint32_t IRCommand;

//void decodeIR(const uint16_t duration) {
//	// add your own protocols here or create your own decoding function
//	// you can overwrite this function in your .ino file
//	decodeAll(duration);
//}

void FuzzyTank::updateIRReceiver()
{
	//uint8_t oldSREG = SREG;
	//cli();
	if (IRProtocol) {
		// print as much as you want in this function
		// see readme to terminate what number is for each protocol
		irReceivePacket.NEC_address.address = IRAddress;
		irReceivePacket.NEC_command.command = IRCommand;
		IRProtocol = 0;
		handleIRPacket();
	}
	//SREG = oldSREG;
}

void IREvent(uint8_t protocol, uint16_t address, uint32_t command) {
	// called when directly received a valid IR signal.
	// do not use Serial inside, it can crash your program!

	// dont update value if blocking is enabled
	if (IRL_BLOCKING && !IRProtocol) {
		// update the values to the newest valid input
		IRProtocol = protocol;
		IRAddress = address;
		IRCommand = command;
	}
}

void FuzzyTank::handleIRPacket()
{
	Serial << F("Byte 1: ") << irReceivePacket.NEC_address.byte1;
	Serial << F("  Byte 2: ") << irReceivePacket.NEC_address.byte2;
	Serial << F("  Byte 3: ") << irReceivePacket.NEC_command.byte3;
	Serial << endl;

	hit();
}

#pragma endregion IR